;;; 
(require (lib "breakpoint.ss"))
; Ein Container: (Fassungsvermoegen (Liste der Stuecke))
; Prioritaetswarteschlange ((Gesamtkosten-1 (Liste der Moeglichkeiten)) (Ges...))

;;; ===== Alternativen =====
; bestimmt die Liste der Moeglichkeiten, das uebergebene Stueck in die
; Container einzufuegen.
; Bei drei Containern gibt es maximal drei Moeglichkeiten, daher eine Liste
; von dann drei Containerlisten.
; Das Problem bei dieser Funktion ist, dass beim Einbauen in den zweiten
; Container der unveraenderte erste Container mit aufgenommen werden muss.
; Dazu werden die Container bei der Bearbeitung in die Liste "vordere"
; aufgenommen. Diese kann dann vorgesetzt werden.
; Die Funktion veraendert nicht die Reihenfolgen.
(define
  (Alternativen Container-Liste Stueck)
  (let loop
      ((Container-Liste Container-Liste)
       (vordere '()) ; der vordere Teil der Container-Liste
       (akku '())) ; die schon erarbeiteten Alternativen
    (cond
      ((null? Container-Liste) (reverse akku))
      ; zu viel?
      ((< (caar Container-Liste) (apply + (cons Stueck (cadar Container-Liste))))
       (loop
        (cdr Container-Liste)
        (cons (car Container-Liste) vordere)
        akku))
      (else
       (loop
        (cdr Container-Liste)
        (cons (car Container-Liste) vordere)
        (cons
         (append
          (reverse vordere)           ; unveraenderter Anfang
          (list
           (list
            (caar Container-Liste)     ; Kapazitaet
            (cons Stueck (cadar Container-Liste)))) ; verlaengerte Liste
           (cdr Container-Liste))    ; Rest
         akku))))))

;;; ===== fuege-hinzu =====
; fuegt eine neue Alternative in die Prioritaetsschlange ein
; Bezeichnungen:
; CoLi : Container-Liste
; GeKo : Gesamtkosten
; Prio : Prioritaetsschlange
(define
  (fuege-hinzu CoLi Geko Prio)
  (cond
    ((null? Prio)
     ; Prio ist noch leer, Liste aufbauen
     (list (list Geko CoLi)))
    ((< Geko (caar Prio)) ; neue Bewertung einfuegen
     (cons (list Geko CoLi) Prio))
    ((= Geko (caar Prio)) ; Position gefunden, nun einfuegen
     (cons
      (fuege-ein CoLi (car Prio))
      (cdr Prio)))
    (else                 ; weiter Psoition suchen
     (cons
      (car Prio)
      (fuege-hinzu CoLi Geko (cdr Prio))))))

;;; ===== fuege-ein =====
; fuegt eine Container-Liste in ihre Teilliste ein
(define
  (fuege-ein CoLi Teil-Liste)
  (cons
   (car Teil-Liste)     ; die Bewertung
   (cons CoLi (cdr Teil-Liste))))

;;; ===== fuege-alle-hinzu =====
; fuegt die neuen Knoten in die Prioritaetsschlange ein
(define
  (fuege-alle-hinzu Alternativen Prio)
  (if
   (null? Alternativen)
   Prio
   (fuege-alle-hinzu
    (cdr Alternativen)
    (fuege-hinzu
     (car Alternativen)
     (Gesamtkosten (car Alternativen))
     Prio))))

;;; ===== Gesamtkosten =====
; Beim Verschnitt duerfen nur hoechstens die restlichen Stuecke
; abgezogen werden. Damit diese Stueckeliste nicht durch die
; Aufrufkette mitgeschleppt werden muss, wird sie hier durch
; Zugriff auf die global definierte Liste "alle-Stuecke" bestimmt.
; Dazu muss gewaehrleistet sein, dass die Stueckeliste immer nur
; Stueck fuer Stueck abgearbeitet wird (immer das naechste).
(define
  (Gesamtkosten alle-Container)
  (let loop
      ((alle-Container alle-Container)
       (Stuecke ())
       (Kosten 0)
       (Verschnitt 0))
;    (bkpt alle-Container)
    (cond
      ((null? alle-Container)
       (+ Kosten (Restkosten Stuecke Verschnitt)))
      ((null? (cadar alle-Container))
       ;noch leer, hierdurch keine Kosten
       (loop (cdr alle-Container) Stuecke Kosten Verschnitt))
      (else
       (loop
         (cdr alle-Container)
         (append (cadar alle-Container) Stuecke)
         (+ Kosten (caar alle-Container))
         (+ Verschnitt (- (caar alle-Container) (apply + (cadar alle-Container)))))))))

;;; ===== Restkosten =====
; Zusaetzliche Restkosten treten nur dann auf, wenn die restlichen
; Stuecke nicht in den Verschnitt passen.
(define
  (Restkosten Stuecke Verschnitt)
  (max
   0
   (- (- (apply + alle-Stuecke) (apply + Stuecke))
      Verschnitt)))
;;; ===== entferne-erste =====
; entfernt das erste Element aus der Prioritaetswarteschlange
(define
  (entferne-erste Prio)
  (cond
    ((null? Prio) #f) ; das geht nicht!
    ((null? (cddar Prio)) ; nur eine mit dem Wert ganz raus!
     (cdr Prio))
    (else
     (cons
      (cons
       (caar Prio)  ; Bewertung
       (cddar Prio))  ; restliche Containerliste mit dem Wert
      (cdr Prio)))))

;;; ===== suche =====
; Das Problem ist, das richtige naechste Stueck zu kennen, da die
; Listen verschieden lang sein koennen.
(define
  (suche Kapazitaeten-Liste Stuecke-Liste)
  (let
    ((Container-Liste (erstelle-CoLi Kapazitaeten-Liste)))
    (let loop
      ((Container-Liste Container-Liste)
       (Prio
        (fuege-alle-hinzu
         (Alternativen
          Container-Liste
          (naechstes-Stueck Container-Liste))
         '())))
      (cond
        ((null? Prio) #f) ; keine Loesung moeglich
        ((alle-drin? Container-Liste) Container-Liste)   ; Loesung!
        (else
         (loop
          (cadar Prio)
          (fuege-alle-hinzu
           (Alternativen (cadar Prio)(naechstes-Stueck Container-Liste))
           (entferne-erste Prio))))))))

;;; ===== erstelle-CoLi =====
; Hilfsfunktion, die zu einer Kapazitaetenliste die Liste der leeren
; Container erstellt.
(define
  (erstelle-CoLi Kapazitaetenliste)
  (cond
    ((null? Kapazitaetenliste) ())
    (else
     (cons
      (list (car Kapazitaetenliste) '())
      (erstelle-CoLi (cdr Kapazitaetenliste))))))

;;; ===== alle-drin? =====
; greift auf die globale Variable alle-Stuecke zu
(define
  (alle-drin? CoLi)
  (let loop
    ((CoLi CoLi)
     (anzahl 0))
  (if
   (null? CoLi)
   (= (length alle-Stuecke) anzahl)
   (loop (cdr CoLi) (+ anzahl (length (cadar CoLi)))))))

;;; ===== naechstes-Stueck =====
; greift auf die globale Variable alle-Stuecke zu
(define
  (naechstes-Stueck CoLi)
  (let loop
    ((CoLi CoLi)
     (anzahl 0))
  (if
    (null? CoLi)
    (list-ref alle-Stuecke anzahl)
    (loop (cdr CoLi) (+ anzahl (length (cadar CoLi)))))))
     
     
;;; ----- Tests -----
(define alle-Stuecke '(30 20 20 20 10))
(define leere-Prio '())
;(Gesamtkosten leere-Container)              ; -> 100
;(Gesamtkosten '((60 (30 20)) (50 (20)) (50 ())))
;(Gesamtkosten '((60 (30)) (50 (20)) (50 (20))))
;(Gesamtkosten '((60 ()) (50 (30 20)) (50 (20))))
;(alle-drin?  '((60 (30)) (50 (20)) (50 (20))))
;(alle-drin?  '((60 (30 20 10)) (50 (20 20)) (50 ())))
;
"erstes Beispiel:"
(define Kapazitaets-Liste-1 '(60 50 50))
(define leere-Container (erstelle-CoLi Kapazitaets-Liste-1))
Kapazitaets-Liste-1
leere-Container
alle-Stuecke
(suche Kapazitaets-Liste-1 alle-Stuecke)
"-----------------"
"zweites Beispiel:"
(define Kapazitaets-Liste-2 '(60 60 50))
(define leere-Container (erstelle-CoLi Kapazitaets-Liste-2))
Kapazitaets-Liste-2
leere-Container
alle-Stuecke
(suche Kapazitaets-Liste-2 alle-Stuecke)
"-----------------"
"drittes Beispiel:"
(define Kapazitaets-Liste-3 '(60 40 50))
(define leere-Container (erstelle-CoLi Kapazitaets-Liste-2))
Kapazitaets-Liste-3
leere-Container
alle-Stuecke
(suche Kapazitaets-Liste-3 alle-Stuecke)
