;;; <A-Stern-iterativ.scm>
(require (lib "breakpoint.ss"))
;;; Angewandt auf ein "kürzeste Wege" - Problem
;;; nach Turau - Algorithmische Graphentheorie (s.263)
; kW-Baum: Baum der kürzesten Wege in einem Graphen
;   von einer Startecke aus. Hier im Graphen G ohne B.
; B: Liste der Ecken, die noch nicht in den kW-Baum eingebaut sind.
; Schätzfunktion ist wieder die Manhattan - Distanz
; Modellierung des Graphen durch Assoziationsliste:
; ((Zeilen-Nummer (Liste der Spalten-Nummern)) ... ) (der unzulässigen)
; da die zulässigen einfach zu berechnen sind.
;;; ===== globale Variable =====
(define
  Hindernisse-Assoziationsliste
  '((1 (2 6))
    (2 (2 3 4 6))
    (3 (2 6 8 9))
    (4 (6 8 9))
    (5 (9))
    (6 (3 4 5 6))
    (7 (3 7))
    (8 (3 7))))
(define
  Hindernisse
  (let loop-1
    ((liste Hindernisse-Assoziationsliste)
     (akku-1 '()))
    (cond
      ((null? liste) akku-1)
      (else
       (loop-1
        (cdr liste)
        (append
         akku-1
         (let loop-2
           ((cdr-liste (cadar liste))
            (akku-2 '()))
           (cond
             ((null? cdr-liste) akku-2)
             (else
              (loop-2
               (cdr cdr-liste)
               (cons (cons (caar liste) (car cdr-liste)) akku-2)))))))))))
(define Start '(9 . 6))
(define Ziel '(1 . 4)) 

; Bildquelle. Turau (s.o.)

;;; ===== Nachfolger =====
; Bestimmt die Nachfolger des ersten Weges
; liste x liste x liste --> neue Wege - Liste
(define
  (Nachfolger Wege besucht Hindernisse)
  (let loop
    ((naechste (Nachbarfelder (caar Wege) besucht Hindernisse))
     (akku (cdr Wege)))
  ;(bkpt "in loop Nachfolger: " Wege #\newline besucht #\newline naechste)
    (cond
      ((null? naechste) akku)
      (else
       (loop
        (cdr naechste)
        (cons (cons (car naechste) (car Wege)) akku))))))

;;; ===== Nachbarfelder =====
; Bestimmt alle zulässigen Nachbarfelder.
; Da bei diesem speziellen Problem mit uniformer Kantenbewertung (1)
; alle Nachbarfelder nächste Nachbarn darstellen, müssen nur einige
; ausgeschlossen werden.
; Paar x liste x liste --> Liste
(define
  (Nachbarfelder Feld besucht Hindernisse)7
  (let loop
    ((alle
      (list
       (cons (sub1 (car Feld)) (cdr Feld))
       (cons (car Feld) (sub1 (cdr Feld)))
       (cons (add1 (car Feld)) (cdr Feld))
       (cons (car Feld) (add1 (cdr Feld)))))
     (akku '()))
    (cond
      ((null? alle) akku)
      ((zulaessig? (car alle) besucht Hindernisse)
       (loop (cdr alle) (cons (car alle) akku)))
      (else
       (loop (cdr alle) akku)))))

;;; ===== zulaessig? =====
; Prüft, ob ein Feld zulässig ist.
; Schon im Baum vorhandene Punkte entfallen, ebenso Hindernisse.
; Außerdem darf der Rand nicht überschritten werden.
; Paar x liste x liste --> boolean
(define
  (zulaessig? Feld besucht Hindernisse)
  (and
   (> (car Feld) 0)
   (> (cdr Feld) 0)
   (< (car Feld) 11)
   (< (cdr Feld) 11)
   (not (member Feld besucht))
   (not (member Feld Hindernisse))))

;;; ===== Distanz =====
; Berechnet die Manhattan - Distanz zweier Felder.
; Paar x Paar --> integer
(define
  (Distanz Feld-1 Feld-2)
  (+
   (if (> (car Feld-1) (car Feld-2))
       (- (car Feld-1) (car Feld-2))
       (- (car Feld-2) (car Feld-1)))
   (if (> (cdr Feld-1) (cdr Feld-2))
       (- (cdr Feld-1) (cdr Feld-2))
       (- (cdr Feld-2) (cdr Feld-1)))))

;;; ===== erstelle-besucht-Liste =====
; Erstellt aus der Wegeliste die besucht - Liste.
; liste --> liste
(define
  (erstelle-besucht-Liste Wege)
  (if
   (null? Wege)
   '()
   (let loop
     ((Wege (cdr Wege))
      (Weg (car Wege))
      (akku '()))
     (cond
       ((and (null? Wege) (null? Weg)) akku)
       ((null? Weg)
        (loop (cdr Wege) (car Wege) akku))
       ((member (car Weg) akku)
        (loop Wege (cdr Weg) akku))
       (else
        (loop Wege (cdr Weg) (cons (car Weg) akku)))))))

;;; ===== A*iter =====
; Die eigentliche Funktion.
(define
  (A*iter Start Ziel)
  (let loop
    ((besucht (erstelle-besucht-Liste '()))
     (Wege (list (list Start)))
     (Grenze (add1 (Distanz Start Ziel))))
    (cond
      ((null? Wege)
       (writeln "Grenze überschritten, Neustart mit " Grenze)
       (loop
        (list Start)
        (list (list Start))
        (add1 Grenze)))
      ((member Ziel (car Wege))
       (writeln)
       (writeln "Am Ziel !")
       (writeln "Lösung :" (reverse (car Wege)))
       (reverse (car Wege)))
      ((> (+ (sub1 (length (car Wege))) (Distanz (caar Wege) Ziel)) Grenze)
       ; Der aktuelle Weg überschreitet die Schranke
       (writeln "kill that cat!" (car Wege))
       (loop besucht (cdr Wege) Grenze))
      (else
       (loop
        (erstelle-besucht-Liste Wege)
        (Nachfolger Wege besucht Hindernisse)
        Grenze)))))

;;; ===== Aufruf =====
(A*iter Start Ziel)