;;;                                <minimal spanning tree.scm>
;;; Spezialproblem, orientiert am 
;;; Problem des Monats von Juli/August 1999 :
;;; Vier Orte liegen in den Eckpunkten eines Quadrates. 
;;; Es soll ein gradliniges Strassennetz aufgebaut werden, 
;;; auf dem man von jedem Ort zu jedem anderen gelangen kann.
;;; Wie sieht ein Strassennetz aus, das eine minimale
;;; Gesamtlaenge hat?
;;; Das Problem weicht in so fern vom minimal spanning tree
;;; ab, als neue Punkte eingefhrt werden drfen.
;;; ==========================================================
;(require (lib "graphics.ss" "graphics") )
;(open-graphics)
;(define w (open-viewport "Fenster" 400 400))

;;; ----- das Quadrat ----------------------------------------
(define a 400)
(define P (make-posn 0 0))
(define Q (make-posn 0 a))
(define R (make-posn a a))
(define S (make-posn a 0))
;                                  >  P    S  <
;
;                                  >  Q    R  <

;;; ----- geometrische Funktionen ----------------------------

;;; ===== Abstand ============================================
; point x point -> distanz
(define
  (Abstand P Q)
  (sqrt (+
   (* (- (posn-x P) (posn-x Q)) (- (posn-x P) (posn-x Q)))
   (* (- (posn-y P) (posn-y Q)) (- (posn-y P) (posn-y Q))))))

;;; ===== Laenge-Netz ========================================
; Berechnet die Laenge des gesamten Netzes aus den Lngen der
; einzelnen Kanten. Dabei sind die Kanten als Liste aus zwei
; Positionen gegeben.
; Netz -> 
(define
  (Laenge-Netz Netz)
  (cond
    ((null? Netz) 0)
    (else
    (+ (Abstand (caar Netz) (cadar Netz))
       (Laenge-Netz (cdr Netz))))))
;;; ===== Beispiel-Netz ======================================
; ein Beispielnetz, bei dem drei Kanten enthalten sind.
(define 
  Beispiel-Netz 
  (list
   (list P Q)
   (list Q R)
   (list R S)))

;;; ===== Problemloesung =====================================
; Weicht man von den 'trivialen Fllen' der unmittelbaren Ver-
; bindungen der Eckpunkte ab, wird die Loesung schwieriger. 
; Allerdings gilt:
; Das Problem ist hochgradig symmetrisch. Alle einzufgenden
; zustzlichen Hilfspunkte mssen also symmetrisch auftreten.

;;; ----- ein weiterer Punkt ---------------------------------
; ein weiterer Punkt kann nur der Mittelpunkt sein. Er braucht
; aber nicht angegeben zu werden, da die Diagonalen sich
; trivialerweise in ihm schneiden.
(define
  einer-mehr-Netz 
  (list
   (list P R)
   (list Q S)))
   
;;; ----- zwei weitere Punkte --------------------------------
; zwei weitere Punkte mssen daher notwendig symmetrisch auf
; der Mittellinie liegen.
;                                  >  P       S  <
;                                       Li Re
;                                  >  Q       R  <

(define Links (make-posn 10 (/ a 2)))
(define Rechts (make-posn (+ 10 (/ a 2)) (/ a 2)))
(define
  zwei-mehr-Netz 
  (list
   (list P Links)
   (list Q Links)
   (list Links Rechts)
   (list Rechts R)
   (list Rechts S)))

;;; ===== Musteraufrufe ======================================
(Laenge-Netz Beispiel-Netz)
(Laenge-Netz einer-mehr-Netz)
(Laenge-Netz zwei-mehr-Netz)
;;; ===== Loesung ============================================
; eine Verbesserung lsst sich mglicherweise durch 
; Verschieben dieser Punkte in x-Richtung erzielen.
; Die Lsung wird mit 'Kochen' gefunden.
; start-wert und hitze aus 0 .. a/2
; abkuehlung aus 0 .. 1
; Der jeweilige optimale x-Wert wird mit einem durch Zufall
; festgelegten Wert verndert. Dabei werden die zulssigen
; Schwankungen (hitze) jeweils mit dem Faktor abkuehlung
; reduziert.
; {1} bestes-x und beste-laenge werden beim Start so gesetzt,
; dass jeder erste Versuch besser sein muss.
; {2} Ist der neue wert nicht besser, wird mit dem alten
; weiter gearbeitet.
(define
  (loese-mit-Kochen start-wert hitze abkuehlung)
  (let
      loop
    ((x start-wert)
     (t hitze)
     (bestes-x -1)                            ;{1}
     (beste-laenge (* 10 a)))
    (letrec
     ((Links (make-posn x (/ a 2)))
      (Rechts (make-posn (- a x) (/ a 2)))
      (Netz 
       (list
        (list P Links)
        (list Q Links)
        (list Links Rechts)
        (list Rechts R)
        (list Rechts S)))
      (laenge
      (Laenge-Netz Netz)))
     (display laenge)(newline)
     (cond
       ((< t 1) bestes-x)
       ((> laenge beste-laenge)                        ;{2}
        (loop 
         (abs (+ bestes-x 
                 (- (random (ex-rund t)) (ex-rund (/ t 2)))))
         (* t abkuehlung)
         bestes-x
         beste-laenge))        
       (else
        (loop
         (abs (+ x (- (random (ex-rund t)) (ex-rund (/ t 2)))))
         (* t abkuehlung)
         x
         laenge))))))
;;; ===== ex-rund ============================================
; Hilfsfunktion zum exakten Runden.
; round : inexact -> inexact
; ex-rund : inexact -> exact
(define (ex-rund x) (inexact->exact (round x)))
;;; ===== Aufruf =============================================
; mit (loese-mit-Kochen start-wert hitze abkuehlung)
(loese-mit-Kochen 10 (/ a 2) 0.95)
(display "Analytische Loesung : ")
(display (/ a 2 (sqrt 3)))
;;; ===== Grafik beenden =====================================
;(close-viewport w)
;(close-graphics)
