;;;                                                  <scheduling.scm>
;;; Problemloesung zum time - scheduling - Problem 
;;; mit Hilfe des A* - Algorithmusses nach
;;; Oliver Grillmeyer : Exploring Computer-Science with Scheme
;;;                     Ab S. 417
;;; >>>>>Das Verfahren:
; Implementiert eine verallgemeinerte A*-Suche auf der Basis einer
; Bestensuche.
; "A*search combines the cost that has already been incured from the
; start state to the current state. Thus a*search is a mix of 
; --> uniform-cost search and
; --> best-first search.

;;; >>>>>Das Problem:
; We are trying to fill time slots during the day with people who are
; available to work at those times. Each person can work a maximum of
; two hours in a day.

;;; ===== a*search ==================================================
; Start-Pfad ist der Pfad zum Ausgangszustand.
; "Nachfolger" ist eine Funktion, welche die naechsten Zustaende
; erzeugt.
(define
  (a*search Ziel-Funktion Start-Pfad Nachfolger g h)
  (best-first-search 
   Ziel-Funktion
   (list Start-Pfad)
   (lambda 
       (Pfad-1 Pfad-2)
     (<= (+ (g Pfad-1) (h Pfad-1)) (+ (g Pfad-2) (h Pfad-2))))
   Nachfolger))

;;; ===== best-first-search =========================================
; Ziel? liefert "true", wenn ein Zielzustand erreicht wurde.
; Auswahlen ist eine Liste von Pfaden, die durchsucht werden sollen.
; Kosten ist eine Funktion, welche die neuen Zustaende ordnet.
; Nachfolger erzeugt die aktuell moeglichen neuen Zustaende.
; Die Funktion liefert den ersten Pfad zurueck, der die Zielfunktion
; erfuellt.
; Die aktuelle Liste der Pfade wird jeweils ausgegeben.
(define
  (best-first-search Ziel? Auswahlen Kosten Nachfolger)
  ;(display Auswahlen)(newline)
  (cond
    ((null? Auswahlen) #f)
    ((Ziel? (first Auswahlen)) (first Auswahlen))
    (else
     (best-first-search
      Ziel?
      (add-paths 
       (rest Auswahlen)
       (make-paths 
	(Nachfolger (first Auswahlen))
	(first Auswahlen))
       Kosten)
      Kosten
      Nachfolger))))

;;; ----- Hilfsfunktionen -------------------------------------------
(define 1+ add1)
;;; ===== subseq ====================================================
; Berechnet die Teilliste ab Startelement bzw. zwischen Start+1 und 
; Ende. Index 1..n
(define 
  subseq
  (lambda
      (liste start . end)
    (cond
      ((null? end) (reverse (list-tail (reverse liste) start)))
    ((< (car end) (length liste))
     (list-tail
      (reverse (list-tail (reverse liste) (- (length liste) (car end))))
      start))
    (else
     (list-tail liste start)))))
;;; ===== position ==================================================
; Liefert die Position eines Elementes in einer Liste oder #f.
(define
  (position element liste)
  (let
      ((ab (member element liste)))
  (if
   ab
   (- (length liste) (length ab))
   #f)))

;;; ===== insert ====================================================
; Liefert eine neue sortierte Liste mit dem eingeordneten Element.
; Das Vergleichspraedikat wird als Parameter �bergeben.
(define
  (insert Element sortierte-Liste Praedikat?)
  (cond
    ((null? sortierte-Liste) (list Element))
    ((Praedikat? Element (first sortierte-Liste))
     (cons Element sortierte-Liste))
    (else
     (cons
      (first sortierte-Liste) 
      (insert Element (rest sortierte-Liste) Praedikat?)))))

;;; ===== add-paths =================================================
; Fuegt die neuen Moeglichkeiten zu den alten sortiert hinzu.
; Das Sortierkriterium liefert die Kostenfunktion.
(define
  (add-paths alte neue Kosten)
  (if 
   (null? neue)
   alte
   (insert
    (first neue)
     (add-paths alte (rest neue) Kosten)
     Kosten)))
;;; ===== make-paths ================================================
; Erhaelt die neuen Zustaende, entfernt leere Listen und erzeugt
; komplette Pfade.
(define
  (make-paths neue aktuell)
  (map
   (lambda (Zustand) (append aktuell (list Zustand)))
   (remove '() neue)))

;;; ===== keep-if ===================================================
; Liefert eine Liste mit allen Elemente, die das Praedikat erfuellen.
(define
  (keep-if Praedikat? Liste)       
    (cond
      ((null? Liste) '())
      ((not
	(Praedikat? (first Liste)))
       (keep-if Praedikat? (rest Liste)))
      (else
       (cons
	(first Liste)
       (keep-if Praedikat? (rest Liste))))))
;;; ===== count =====================================================
; Liefert die Anzahl, wie oft ein Element in einer Liste auftritt.
(define
  (count Element Liste)
  (cond
    ((null? Liste) 0)
    ((equal? Element (first Liste))
     (1+ (count Element (rest Liste))))
    (else
     (count Element (rest Liste)))))

;;; ===== Problem - abhaengige Funktionen ===========================
(define
  (possible-choices Zeit Pfad)
  (map 
   first
   (keep-if
    (lambda (Zeiten-von-Person)
      (and
       (member Zeit (rest Zeiten-von-Person))
       (< (count (first Zeiten-von-Person) Pfad) 2)))
    people)))

;;; ===== Daten =====================================================
(define times-to-fill '(8 9 10 11 12 13 14 15 16 17))

(define
  people
  '((Anna 8 9 10 11 12)
    (Bertha 12 13 14 16 17)
    (Christiane 8 9 10 15 16 17)
    (Dorothea 11 13 14 15)
    (Elisabeth 9 10 11 14 15 16)))

;;; ===== Aufrufe ===================================================
; Das besondere Problem bei der a*-Suche ist die Bewertungsfunktion.
; Dabei ist die Wahl von f selbstverst�ndlich: Sie gibt die Kosten
; bis zum aktuellen Knoten an. h ist allerdings eine Sch�tzfunktion
; f�r die Reskosten, die bestimmte mathematische Bedingungen erf�llen
; sollte. Hier wird zun�chst eine sehr einfache gew�hlt: 
;   h  <-- Zahl der verbleibenden Zeiten.
; ----- der Aufruf --------------------------------------------------
;;;(a*search Ziel-Funktion Start-Pfad Nachfolger g h)
; Ziel-Funktion: alle besetzt.
; Start-Pfad   : leer.
; Nachfolger   : Alle Personen, die noch frei sind.
; g            : L�nge des aktuellen Pfades.
; h            : L�nge der noch zu f�llenden Zeiten.
;(a*search 
; (lambda (path) (= (length path) (length times-to-fill)))
; '()
; (lambda (path) 
;   (possible-choices 
;    (list-ref times-to-fill (length path))
;    path))
; length
; (lambda (path) (- (length times-to-fill) (length path)
;;;              ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
; Die Frage ist, ob sie die Restkosten richtig sch�tzt. Antwort:nein!
; Die erstaunliche zu fordernde Eigenschaft der Restkostenfunktion
; ist, dass sie die Kosten nicht zu >>>hoch<<< einsch�tzen darf. Das
; tut die o.a. aber. Daher:

;;; ===== zweite Variante ===========================================
; (S.424) For the function h we can compute the sum of the people who
; will work in the remaining time slots to fill. Since there can be
; many people who can work at a given time slot, we can compute the 
; costs for each available person and use the minimum costs. Finally
; we sum all the remainig costs for each remainig time...
; ---------------- Hilfsvariante von min ----------------------------
; Um zu verhindern, dass min auf eine leere Liste angewandt wird.
(define maxzahl (* (length people) (length times-to-fill)))
(define mini min)
(define
  min 
  (lambda
      Parameter
    (if (null? Parameter) maxzahl (apply mini Parameter))))
; ----- der Aufruf: -------------------------------------------------
;;;(a*search Ziel-Funktion Start-Pfad Nachfolger g h)
; Zielfunktion: Wie oben, alle besetzt.
; Start-Pfad  : Wie oben, leer.
; Nachfolger  : Wie oben, alle Personen, die noch frei sind.
; g           : Addiere alle ...
; h           : Addiere alle ...
(a*search
  (lambda (path) (= (length path) (length times-to-fill)))
  '()
  (lambda (path)
    (possible-choices (list-ref times-to-fill (length path)) path))
  (lambda (path)
    (apply +
           (map (lambda (person time)
                  (position time (cdr (assoc person people))))
                path (subseq times-to-fill 0 (length path)))))
  (lambda (path) 
    (apply +
      (map (lambda (time)
             (apply min
               (map (lambda (person)
                      (position time (cdr (assoc person people))) )
                 (possible-choices time path))) )
        (subseq times-to-fill (length path))))))

;;; liefert:
;;;(anna elisabeth anna dorothea bertha
;;; dorothea elisabeth christiane bertha christiane)
;;; [Sieht man da den Vorteil? Ich nicht!]
;;; ===== letzte Variante ===========================================
;;;(a*search Ziel-Funktion Start-Pfad Nachfolger g h)
;;; s.o., aber min muss hierf�r wieder auskommentiert werden!
;(a*search
;  (lambda (path) (= (length path) (length times-to-fill)))
;  '()
;  (lambda (path)
;    (possible-choices (list-ref times-to-fill (length path)) path))
;  (lambda (path)
;    (apply +
;      (map (lambda (person time)
;             (position time (cdr (assoc person people))))
;        path (subseq times-to-fill 0 (length path)))))
;  (lambda (path) 
;    (apply +
;      (map (lambda (time)
;             (if (not (null? (possible-choices time path)))
;                 (apply min
;                   (map (lambda (person)
;                          (position time
;                            (cdr (assoc person people))))
;                     (possible-choices time path)))
;                 1000))  ; return a large cost
;        (subseq times-to-fill (length path))))))
;;; liefert:
;;;(anna elisabeth anna dorothea bertha
;;; dorothea elisabeth christiane bertha christiane)
;;; [Sieht man da den Vorteil? Ich nicht!]
;;; ===== ENDE ======================================================