;;; <rucksack mit breitensuche.scm>
(require (lib "breakpoint.ss")) ; bei Version 206 ber module - Deklaration
;;; Eine einfache Lsung des Rucksackproblems mit Breitensuche
;;; Die Funktion Rucksack-packen bildet eine Aufrufhlle fr die 
;;; eigentliche Breitensuchefunktion. Dadurch braucht der Benutzer
;;; nicht beim Aufruf die leere Liste fr den akku zu bergeben.
;;; Auerdem ergibt sich der Vorteil, dass der Parameter Maximal-Gewicht
;;; fr die innere Funktion global ist und daher nicht in der
;;; Aufrufkette mit geschleppt werden muss.

;;; ===== expandiere =====
; Teilrucksack x Gewicht x Maximum ---> dei beiden neuen Teilruckscke (mit und ohne)
(define
  (expandiere Teilrucksack Gewicht Maximum)
  (append
   (list 
    Teilrucksack
    (if
     (> (apply + (cons Gewicht Teilrucksack)) Maximum)
     ()
     (cons Gewicht Teilrucksack)))))

;;; ===== expandiere-alle =====
; Teilrucksackliste x Gewicht x Maximum ---> neue Teilrucksackliste
(define
  (expandiere-alle Teilrucksackliste Gewicht Maximum)
  (cond
    ((null? Teilrucksackliste) '())
    (else
     (append
      (expandiere (car Teilrucksackliste) Gewicht Maximum)
      (expandiere-alle (cdr Teilrucksackliste) Gewicht Maximum)))))

;;; ===== Loesung? =====
; Teilrucksackliste x  Maximum ---> #f / Loesungsliste
(define
  (Loesung? Teilrucksackliste Maximum)
  (cond
    ((null? Teilrucksackliste) #f)
    ((= (apply + (car Teilrucksackliste)) Maximum)
     (car Teilrucksackliste))
    (else
     (Loesung? (cdr Teilrucksackliste) Maximum))))
     

;;; ===== Rucksack-packen =====
(define
  (Rucksack-packen Maximal-Gewicht Teilgewichte)
  (let b-s
    ((queue (expandiere-alle '(()) (car Teilgewichte) Maximal-Gewicht))
     (Teilgewichte (cdr Teilgewichte)))
    ;(bkpt "Schlange: " queue #\newline "restliche Gewichte: " Teilgewichte)
    (cond
      ((Loesung? queue Maximal-Gewicht))
      ((null? Teilgewichte) #f)
      (else
       (b-s 
        (expandiere-alle queue (car Teilgewichte) Maximal-Gewicht)
        (cdr Teilgewichte))))))

(define Teilgewichte-a '(1 1 40 2 3 5 15 15 20 40))
(Rucksack-packen 101 Teilgewichte-a)
;(Rucksack-packen 102 Teilgewichte-a)
;(Rucksack-packen 103 Teilgewichte-a)
;(Rucksack-packen 104 Teilgewichte-a)
;(Rucksack-packen 105 Teilgewichte-a)
;
;(define Teilgewichte-b '(1 2 5 9 20 50 90 181 400 900))
;(Rucksack-packen 101 Teilgewichte-b)
;(Rucksack-packen 102 Teilgewichte-b)
;(Rucksack-packen 103 Teilgewichte-b)
;(Rucksack-packen 104 Teilgewichte-b)
;(Rucksack-packen 105 Teilgewichte-b)
