;;; Rucksack mit greedy.scm
(require (lib "breakpoint.ss"))
; Erster Lsungsansatz zum [Rucksackproblem] linearen Container - 
; Lade - Problem mit einer gierigen Strategie. Die Teilgewichte sollten
; der Grsse nach geordnet vorliegen. Anderenfalls muss man sie so
; sortieren.
; Es ist zu klren, ob Satisfizierung oder Optimierung gewnscht wird.
;;; ===== Rucksack-packen ============================================
; Bei diesem Algorithmus ist eine einfache Rekursion mglich. Sie ist
; aber wegen der Abbruchbedingungen unhandlich. Also endrekursiv mit
; Akku {Ergebnis}
; Gesamtgewicht x sortierte Liste der mglichen Teilgewichte x ()
;              --> Liste der optimalen TG
; 1. Fall: genau erfllen (Satisfizierung)
(define
  (Rucksack-packen Gewicht Teilgewichte Ergebnis)
  (cond
    ((zero? Gewicht)          ; Lsung gefunden !
     Ergebnis)
    ((null? Teilgewichte)     ; nicht lsbar
     #f)
    ((< Gewicht (car Teilgewichte))        ; erster geht nicht
     (Rucksack-packen Gewicht (cdr Teilgewichte) Ergebnis))
    (else              ; einpacken und weiter
     (Rucksack-packen
      (- Gewicht (car Teilgewichte))
      (cdr Teilgewichte)
      (cons (car Teilgewichte) Ergebnis)))))
     
; 2. Fall: optimieren
(define
  (Rucksack-packen Gewicht Teilgewichte Ergebnis)
  (cond
    ((zero? Gewicht)          ; Lsung gefunden !
     Ergebnis)
    ((null? Teilgewichte)     ; nicht voll erfllt
     (cons "nicht exakt" Ergebnis))
    ((< Gewicht (car Teilgewichte))        ; erster geht nicht
     (Rucksack-packen Gewicht (cdr Teilgewichte) Ergebnis))
    (else              ; einpacken und weiter
     (Rucksack-packen
      (- Gewicht (car Teilgewichte))
      (cdr Teilgewichte)
      (cons (car Teilgewichte) Ergebnis)))))
     
; Will man den Ablauf verfolgen, setzt man unmittelbar vor dem cond
; den Aufruf   (bkpt Gewicht Teilgewichte Ergebnis)    ein.
     

(define Teilgewichte-a '(30 30 30 20 20 20 15 15 5 3 2 1 1))
(Rucksack-packen 101 Teilgewichte-a ())
;(Rucksack-packen 102 Teilgewichte-a ())
;(Rucksack-packen 103 Teilgewichte-a ())
;(Rucksack-packen 104 Teilgewichte-a ())
;(Rucksack-packen 105 Teilgewichte-a ())
;
;(define Teilgewichte-b '(900 400 181 90 50 20 9 5 2 1))
;(Rucksack-packen 101 Teilgewichte-b ())
;(Rucksack-packen 102 Teilgewichte-b ())
;(Rucksack-packen 103 Teilgewichte-b ())
;(Rucksack-packen 104 Teilgewichte-b ())
;(Rucksack-packen 105 Teilgewichte-b ())
