;;; Rucksack mit tiefensuche.scm
;;; Eine Lsung des Rucksackproblems mit Tiefensuche
;;; Die Funktion Rucksack-packen bildet eine Aufrufhlle fr die 
;;; eigentliche Tiefensuchefunktion. Dadurch braucht der Benutzer
;;; nicht beim Aufruf die leere Liste fr den akku zu bergeben.
;;; Auerdem ergibt sich der Vorteil, dass der Parameter Gewicht
;;; fr die innere Funktion global ist und daher nicht in der
;;; Aufrufkette (eigentlich -Baum) mit geschleppt werden muss.
;;; Die Alternativen werden hier - wie bei der einfachen Funktion -
;;; nicht bestimmt, sondern nur beim Neuaufruf bergeben.

(define
  (Rucksack-packen Maximal-Gewicht Teilgewichte)
  (let t-s
    ((Teilgewichte Teilgewichte)
     (akku '()))
    (cond
      ((null? Teilgewichte) #f)
      ((Ziel? akku Maximal-Gewicht)
       akku)
      ; In einem Baum kein Zyklus, daher sofort:
      ; erster Fall: einpacken!
      ((t-s (cdr Teilgewichte) (cons (car Teilgewichte) akku)))
      (else
       ; die einzige Alternative: nicht einpacken!
       (t-s (cdr Teilgewichte) akku)))))

;;; ===== Ziel? =====
; Gewichte x Maximal-Gewicht --> #f / #t
(define
  (Ziel? Gewichte Maximal-Gewicht)
  (= Maximal-Gewicht (apply + Gewichte)))


(define Teilgewichte-a '(1 1 40 2 3 5 15 15 20 40))
(Rucksack-packen 101 Teilgewichte-a)
(Rucksack-packen 102 Teilgewichte-a)
(Rucksack-packen 103 Teilgewichte-a)
(Rucksack-packen 104 Teilgewichte-a)
(Rucksack-packen 105 Teilgewichte-a)

(define Teilgewichte-b '(1 2 5 9 20 50 90 181 400 900))
(Rucksack-packen 101 Teilgewichte-b)
(Rucksack-packen 102 Teilgewichte-b)
(Rucksack-packen 103 Teilgewichte-b)
(Rucksack-packen 104 Teilgewichte-b)
(Rucksack-packen 105 Teilgewichte-b)
