;;;                                     "rucksack-gierig-entwicklung.scm"
(require (lib "breakpoint.ss"))
; ----- Liste der vorhandenen Stuecke -----
(define stuecke '(30 30 30 20 20 15 5 2 1))

;;; ===== ANFANG HILFSFUNKTIONEN =====
; ----- Praedikate -----
; Fuer das Verstaendnis ist vermutlich einfacher, diese Prädikate auszugliedern:
; ziel? prueft, ob alle Werte zusammen den Zielwert ergeben
(define (ziel? container zielwert)    (= zielwert (apply + container)))
; zu-voll? prueft, ob alle Werte zusammen den Zielwert ueberschreiten
(define (zu-voll? container zielwert) (< zielwert (apply + container)))

; ----- Hilfsfunktionen fuer bessere Lesbarkeit -----
; Die Restliste einer Liste nach dem Kopf ist natuerlich einfach nur (cdr ...)
(define (restliste liste) (cdr liste))
; Das erste Element entsprechend (car ...)
(define (erstes liste) (car liste))
; Man fuellt ein Stueck ein mit (cons ... ...)
(define (hinein element liste) (cons element liste))
;;; ===== ENDE HILFSFUNKTIONEN =====

;;; ----- Funktionskopf -----
; 1. Schritt: Bestimmung des Funktionskopfes
(define
  (suche stuecke container zielwert)
   'wert-undefiniert)
; Testaufrufe:
(writeln "Schritt 1: ")
(suche stuecke '() 105)
(suche stuecke '() 100)
(suche stuecke '() 200)

;;; ----- am Ziel? -----
; 2. Schritt: ich bin am Ziel, der Container ist exakt gefuellt.
; Hier ist also eine Bedingung zu überprüfen. --> if
(define
  (suche stuecke container zielwert)
  (if
   (ziel? container zielwert)                     ; Bedingung
   container                                      ; ja-Fall: container zurueckgeben
   'wert-undefiniert)                             ; nein-Fall
  )

; Testaufrufe:
(writeln "Schritt 2: ")
(suche '(5 2 1) '(30 30 30 15) 105)
(suche '(5 2 1) '(30 30 30 15) 100)

;;; ----- am "Ende", aber nicht am Ziel -----
; 3. Schritt: es geht nicht weiter!
; Alle Stuecke wurden versucht, ohne dass sich der Container passend fuellen ließ.
; Hier ist also wieder eine Bedingung zu überprüfen. Wir könnten ein weiters if
; verwenden, da Scheme aber eine sehr komfortable Mehrfachverzweigung bereitstellt,
; verwenden wir diese: --> cond
(define
  (suche stuecke container zielwert)
  (cond
    [(ziel? container zielwert)                ; 1.Bedingung
     container]                                ; ja-Fall
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [else                                      ; und sonst?
     'wert-undefiniert]                        ; der "sonst"-Fall
    )
  )
; Testaufrufe:
(writeln "Schritt 3: ")
(suche '(5 2 1) '(30 30 30 15) 105)
(suche '(5 2 1) '(30 30 30 15) 100)
(suche '() '(30 30 30) 100)

;;; ----- zu voll -----
; 4. Schritt: das aktuelle Stueck passt nicht hinein.
; es wird weitergesucht ohne dieses Stueck.
(define
  (suche stuecke container zielwert)
  (cond
    [(ziel? container zielwert)                ; 1.Bedingung
     container]                                ; ja-Fall
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [(zu-voll?                                 ; 3.Bedingung
      (hinein (erstes stuecke) container)      ; "probieren"
      zielwert)
     (suche                                    ; weiter suchen ...
      (restliste stuecke)                      ; ... ohne das erste Element ...
      container                                ; ... mit der bisherigen Fuellung
      zielwert)]                               ; ... und dem selben Zielwert.
    [else                                      ; und sonst?
     'wert-undefiniert]                        ; der "sonst"-Fall
    )
  )
; Testaufrufe:
(writeln "Schritt 4: ")
(suche '(20 15 5 2 1) '(30 30 30) 105)
(suche '() '(30 30 30) 100)
(suche '(15) '(30 30 30) 100)

;;; ----- der "sonst"-Fall -----
; 5. Schritt: das aktuelle Stueck passt hinein.
; es gibt nur noch diese Möglichkeit. Es wird weitergesucht mit diesem Stueck
; im Container.
(define
  (suche stuecke container zielwert)
  (cond
    [(ziel? container zielwert)                ; 1.Bedingung
     container]                                ; ja-Fall
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [(zu-voll?                                 ; 3.Bedingung
      (hinein (erstes stuecke) container)      ; "probieren"
      zielwert)
     (suche                                    ; weiter suchen ...
      (restliste stuecke)                      ; ... ohne das erste Element ...
      container                                ; ... mit der bisherigen Fuellung
      zielwert)]                               ; ... und dem selben Zielwert.
    [else                                      ; und sonst?
     (suche                                    ; weiter suchen ...
      (restliste stuecke)                      ; ... ohne das erste Element ...
      (hinein (erstes stuecke) container)      ; ... mit der neuen Fuellung
      zielwert)]                               ; ... und dem selben Zielwert.
    )
  )
; Testaufrufe:
(writeln "Schritt 4: ")
(suche stuecke '() 105)
(suche stuecke '() 100)
(suche stuecke '() 200)
