;;;                                   "rucksack-ts-entwicklung.scm"
;;; Demonstration der Entwicklung des Programmes zur Tiefensuche am Rucksackproblem
(require (lib "breakpoint.ss"))
; ----- Liste der vorhandenen Stuecke -----
(define stuecke '(30 30 30 20 20 15 5 2 1))

;;; ===== ANFANG HILFSFUNKTIONEN =====
; ----- Praedikate -----
; Fuer das Verstaendnis ist vermutlich einfacher, diese Prädikate auszugliedern:
; ziel? prueft, ob alle Werte zusammen den Zielwert ergeben
(define (ziel? container zielwert)    (= zielwert (apply + container)))
; zu-voll? prueft, ob alle Werte zusammen den Zielwert ueberschreiten
(define (zu-voll? container zielwert) (< zielwert (apply + container)))

; ----- Hilfsfunktionen fuer bessere Lesbarkeit -----
; Die Restliste einer Liste nach dem Kopf ist natuerlich einfach nur (cdr ...)
(define (restliste liste) (cdr liste))
; Das erste Element entsprechend (car ...)
(define (erstes liste) (car liste))
; Man fuellt ein Stueck ein mit (cons ... ...)
(define (hinein element liste) (cons element liste))
;;; ===== ENDE HILFSFUNKTIONEN =====

;;; ----- Funktionskopf -----
; 1. Schritt: Bestimmung des Funktionskopfes
(define
  (tiefensuche stuecke container zielwert)
  'wert-undefiniert)
; Testaufrufe:
(writeln "Schritt 1: ")
(tiefensuche stuecke () 105)
(tiefensuche stuecke () 100)

;;; ----- am Ziel? -----
; 2. Schritt: ich bin am Ziel, der Container ist exakt gefuellt.
; Hier ist also eine Bedingung zu überprüfen. --> if
(define
  (tiefensuche stuecke container zielwert)
  (if
   (ziel? container zielwert)                     ; Bedingung
   container                                      ; ja-Fall: container zurueckgeben
   'wert-undefiniert)                             ; nein-Fall
  )

; Testaufrufe:
(writeln "Schritt 2: ")
(tiefensuche '(5 2 1) '(30 30 30 15) 105)
(tiefensuche '(15 5 2 1) '(30 30 30) 100)

;;; ----- am "Ende", aber nicht am Ziel -----
; 3. Schritt: es geht nicht weiter!
; Alle Stuecke wurden versucht, ohne dass sich der Container passend fuellen ließ.
; Hier ist also wieder eine Bedingung zu überprüfen. Wir könnten ein weiters if
; verwenden, da Scheme aber eine sehr komfortable Mehrfachverzweigung bereitstellt,
; verwenden wir diese: --> cond
(define
  (tiefensuche stuecke container zielwert)
  (cond
    [(ziel? container zielwert)                ; 1.Bedingung
     container]                                ; ja-Fall
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [else                                      ; und sonst?
     'wert-undefiniert]                        ; der "sonst"-Fall
    )
  )
; Testaufrufe:
(writeln "Schritt 3: ")
(tiefensuche '(5 2 1) '(30 30 30 15) 105)
(tiefensuche '(5 2 1) '(30 30 30 15) 100)
(tiefensuche '() '(30 30 30 15) 100)

;;; ----- Rekusionssteuerung -----
; 4. Schritt: versuche es mit dem ersten
; Es gibt keine weiteren besonderen Fälle.
; Die Funktion muss nun um die Steuerung der Rekursion erweitert werden.
; Sie muss ...
;     ... versuchen das aktuelle Stueck einzufuellen und(!)
;     ... wenn das nicht erfolgreich ist, versuchen wir es ohne das aktuelle Stueck.
(define
  (tiefensuche stuecke container zielwert)
  (cond
    [(ziel? container zielwert)                ; 1.Bedingung
     container]                                ; ja-Fall
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [(tiefensuche (cdr stuecke) (cons (car stuecke) container) zielwert)]
    ;;; Eine der besonderen Möglichkeiten von cond: Bedingung=Wert
    ;;; Die Bedingung hat einen Wert, der zurückgegeben werden kann.
    ;;; Bei erfolglosem Versuch brauchen wir dann hier den Wert #f,
    ;;; damit zum else gegangen wird.
    [else                                             ; und sonst?
     (tiefensuche (cdr stuecke) container zielwert)]  ; der "sonst"-Fall
    )
  )

; Testaufrufe:
(writeln "Schritt 4: ")
(tiefensuche stuecke '() 105)
(tiefensuche stuecke '() 100)
(tiefensuche stuecke '() 103)
(tiefensuche stuecke '() 500)


;;; ----- Vervollständigung der Werte -----
; 5. Schritt: Vervollständigung der Rekusionssteuerung
; Wenn es nicht geklappt hat, soll das Programm zu den Alternativen zurückspringen.
; Das geschieht, wenn es auf der unteren Stufe ein #f zurückbekommt.
; Dort wo noch 'Wert-noch-unklar steht, muss also #f stehen.
(define
  (tiefensuche stuecke container zielwert)
  (cond
    [(ziel? container zielwert)                ; 1.Bedingung
     container]                                ; ja-Fall
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [(tiefensuche (cdr stuecke) (cons (car stuecke) container) zielwert)]
    [else                                             ; und sonst?
     (tiefensuche (cdr stuecke) container zielwert)]  ; der "sonst"-Fall
    )
  )

; Testaufrufe:
(writeln "Schritt 5: ")
(tiefensuche stuecke '() 105)
(tiefensuche stuecke '() 100)
(tiefensuche stuecke '() 103)
(tiefensuche stuecke '() 500)

;;; ----- Vervollständigung der Ausgabe -----
; 6. Schritt: Wir wollen das Ergebnis sehen!
; Das Programm muss also den Container ausgeben.
(define
  (tiefensuche stuecke container zielwert)
  (cond
    [(ziel? container zielwert)                ; 1.Bedingung
     container]                                ; ja-Fall
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [(tiefensuche (cdr stuecke) (cons (car stuecke) container) zielwert)]
    [else                                             ; und sonst?
     (tiefensuche (cdr stuecke) container zielwert)]  ; der "sonst"-Fall
    )
  )

; Testaufrufe:
(writeln "Schritt 6: ")
(tiefensuche stuecke '() 105)
(tiefensuche stuecke '() 100)
(tiefensuche stuecke '() 103)
(tiefensuche stuecke '() 500)
