;;;                                   "rucksack-ts-entwicklung.scm"
;;; Demonstration der Entwicklung des Programmes zur Tiefensuche am Rucksackproblem
(require (lib "breakpoint.ss"))
; ----- Liste der vorhandenen Stuecke -----
(define stuecke '(30 30 30 20 20 15 5 2 1))

;;; ===== ANFANG HILFSFUNKTIONEN =====
; ----- Praedikate -----
; Fuer das Verstaendnis ist vermutlich einfacher, diese Prädikate auszugliedern:
; ziel? prueft, ob alle Werte zusammen den Zielwert ergeben
(define (ziel? container zielwert)    (= zielwert (apply + container)))
; zu-voll? prueft, ob alle Werte zusammen den Zielwert ueberschreiten
(define (zu-voll? container zielwert) (< zielwert (apply + container)))

; ----- Hilfsfunktionen fuer bessere Lesbarkeit -----
; Die Restliste einer Liste nach dem Kopf ist natuerlich einfach nur (cdr ...)
(define (restliste liste) (cdr liste))
; Das erste Element entsprechend (car ...)
(define (erstes liste) (car liste))
; Man fuellt ein Stueck ein mit (cons ... ...)
(define (hinein element liste) (cons element liste))
;;; ===== ENDE HILFSFUNKTIONEN =====

;;; ----- Funktionskopf -----
; 1. Schritt: Bestimmung des Funktionskopfes
(define
  (tiefensuche stuecke container zielwert)
  'wert-undefiniert)
; Testaufrufe:
(writeln "Schritt 1: ")
(tiefensuche stuecke () 105)
(tiefensuche stuecke () 100)

;;; ----- am Ziel? -----
; 2. Schritt: ich bin am Ziel, der Container ist exakt gefuellt.
; Hier ist also eine Bedingung zu überprüfen. --> if
(define
  (tiefensuche stuecke container zielwert)
  (if
   (ziel? container zielwert)                     ; Bedingung
   container                                      ; ja-Fall: container zurueckgeben
   'wert-undefiniert)                             ; nein-Fall
  )

; Testaufrufe:
(writeln "Schritt 2: ")
(tiefensuche '(5 2 1) '(30 30 30 15) 105)
(tiefensuche '(15 5 2 1) '(30 30 30) 100)

;;; ----- am "Ende", aber nicht am Ziel -----
; 3. Schritt: es geht nicht weiter!
; Alle Stuecke wurden versucht, ohne dass sich der Container passend fuellen ließ.
; Hier ist also wieder eine Bedingung zu überprüfen. Wir könnten ein weiters if
; verwenden, da Scheme aber eine sehr komfortable Mehrfachverzweigung bereitstellt,
; verwenden wir diese: --> cond
(define
  (tiefensuche stuecke container zielwert)
  (cond
    [(ziel? container zielwert)                ; 1.Bedingung
     container]                                ; ja-Fall
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [else                                      ; und sonst?
     'wert-undefiniert]                        ; der "sonst"-Fall
    )
  )
; Testaufrufe:
(writeln "Schritt 3: ")
(tiefensuche '(5 2 1) '(30 30 30 15) 105)
(tiefensuche '(5 2 1) '(30 30 30 15) 100)
(tiefensuche '() '(30 30 30 15) 100)

;;; ----- zu voll -----
; 4. Schritt: das aktuelle Stueck passt nicht hinein.
; es wird weitergesucht ohne dieses Stueck.
(define
  (tiefensuche stuecke container zielwert)
  (cond
    [(ziel? container zielwert)                ; 1.Bedingung
     container]                                ; ja-Fall
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [(zu-voll?                                 ; 3.Bedingung
      (hinein (erstes stuecke) container)      ; "probieren"
      zielwert)
     (tiefensuche                              ; weiter suchen ...
      (restliste stuecke)                      ; ... ohne das erste Element ...
      container                                ; ... mit der bisherigen Fuellung
      zielwert)]                               ; ... und dem selben Zielwert.
    [else                                             ; und sonst?
     'wert-undefiniert]                        ; der "sonst"-Fall
    )
  )

; Testaufrufe:
(writeln "Schritt 4: ")
(tiefensuche stuecke '() 105)
(tiefensuche stuecke '() 100)
(tiefensuche stuecke '() 103)
(tiefensuche stuecke '() 500)


;;; ----- der "sonst"-Fall ??? -----
; 5. Schritt: das aktuelle Stueck passt hinein.
; es gibt nur noch diese Möglichkeit. Es wird weitergesucht mit diesem Stueck
; im Container.
(define
  (tiefensuche stuecke container zielwert)
  (cond
    [(ziel? container zielwert)                ; 1.Bedingung
     container]                                ; ja-Fall
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [(zu-voll?                                 ; 3.Bedingung
      (hinein (erstes stuecke) container)      ; "probieren"
      zielwert)
     (tiefensuche                              ; weiter suchen ...
      (restliste stuecke)                      ; ... ohne das erste Element ...
      container                                ; ... mit der bisherigen Fuellung
      zielwert)]                               ; ... und dem selben Zielwert.
    [else                                      ; und sonst?
     (tiefensuche                              ; weiter suchen ...
      (restliste stuecke)                      ; ... ohne das erste Element ...
      (hinein (erstes stuecke) container)      ; ... mit der neuen Fuellung
      zielwert)]                               ; ... und dem selben Zielwert.
    )
  )

; Testaufrufe:
(writeln "Schritt 5: ")
(tiefensuche stuecke '() 105)
(tiefensuche stuecke '() 100)
(tiefensuche stuecke '() 103)
(tiefensuche stuecke '() 500)

;;; ----- das war es noch nicht ! -----
; 6. Schritt: das Stueck einzupacken fuehrte ggf. in eine Sackgasse.
; wir muessen es auch ohne dies Stueck versuchen.
; Dazu muss der Aufruf mit dem Stueck (in die Tiefe) wie eine Bedingung behandelt
; werden. Wie das z.B. geht, s.u.
(define
  (tiefensuche stuecke container zielwert)
  (cond
    [(ziel? container zielwert)                ; 1.Bedingung
     container]                                ; ja-Fall
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [(zu-voll?                                 ; 3.Bedingung
      (hinein (erstes stuecke) container)      ; "probieren"
      zielwert)
     (tiefensuche                              ; weiter suchen ...
      (restliste stuecke)                      ; ... ohne das erste Element ...
      container                                ; ... mit der bisherigen Fuellung
      zielwert)]                               ; ... und dem selben Zielwert.
     [(tiefensuche                             ; weiter suchen ...
      (restliste stuecke)                      ; ... ohne das erste Element ...
      (hinein (erstes stuecke) container)      ; ... mit der neuen Fuellung
      zielwert)]                               ; ... und dem selben Zielwert.
    ;;; Eine der besonderen Moeglichkeiten von cond: Bedingung=Wert
    ;;; Die Bedingung hat einen Wert, der zurückgegeben werden kann.
    ;;; Bei erfolglosem Versuch brauchen wir dann hier den Wert #f,
    ;;; damit zum else gegangen wird.
    [else                                      ; und sonst?
     (tiefensuche                              ; kennen wir schon:  weiter suchen ...
      (restliste stuecke)                      ; ... ohne das erste Element ...
      container                                ; ... mit der bisherigen Fuellung
      zielwert)]                               ; ... und dem selben Zielwert.
    )
  )

; Testaufrufe:
(writeln "Schritt 6: ")
(tiefensuche stuecke '() 105)
(tiefensuche stuecke '() 100)
(tiefensuche stuecke '() 103)
(tiefensuche stuecke '() 500)


;;; ----- Vervollständigung der Ausgabe -----
; 7. Schritt: kommentierte Ausgabe!
(define
  (tiefensuche stuecke container zielwert)
  (cond
    [(ziel? container zielwert)                ; 1.Bedingung
     (writeln "Fuellung " zielwert " erreicht mit Container " container)
     container]                                ; ja-Fall
    [(null? stuecke)                           ; 2.Bedingung
     #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
    [(zu-voll?                                 ; 3.Bedingung
      (hinein (erstes stuecke) container)      ; "probieren"
      zielwert)
     (tiefensuche                              ; weiter suchen ...
      (restliste stuecke)                      ; ... ohne das erste Element ...
      container                                ; ... mit der bisherigen Fuellung
      zielwert)]                               ; ... und dem selben Zielwert.
     [(tiefensuche                             ; weiter suchen ...
      (restliste stuecke)                      ; ... ohne das erste Element ...
      (hinein (erstes stuecke) container)      ; ... mit der neuen Fuellung
      zielwert)]                               ; ... und dem selben Zielwert.
    ;;; Eine der besonderen Moeglichkeiten von cond: Bedingung=Wert
    ;;; Die Bedingung hat einen Wert, der zurückgegeben werden kann.
    ;;; Bei erfolglosem Versuch brauchen wir dann hier den Wert #f,
    ;;; damit zum else gegangen wird.
    [else                                      ; und sonst?
     (tiefensuche                              ; kennen wir schon:  weiter suchen ...
      (restliste stuecke)                      ; ... ohne das erste Element ...
      container                                ; ... mit der bisherigen Fuellung
      zielwert)]                               ; ... und dem selben Zielwert.
    )
  )

; Testaufrufe:
(writeln "Schritt 7: ")
(tiefensuche stuecke '() 105)
(tiefensuche stuecke '() 100)
(tiefensuche stuecke '() 103)
(tiefensuche stuecke '() 500)
