Programm ist unvollständig
;;; Demonstration der Entwicklung des Programmes zur Tiefensuche am Rucksackproblem
;;; hier fuer viele Rucksaecke, also muss eine Container-Liste verwendet werden.
(require (lib "breakpoint.ss"))
; ----- Liste der vorhandenen Stuecke -----
(define stuecke '(40 30 30 30 25 25 25 20 20 20 15 10 10 10 10 5))

; ----- die Container -----
; Da es nun nicht nur einen Container gibt, müssen wir eine Liste von Containern
; verwenden, die wiederum selbst aus Listen bestehen,
; die am Kopf die Kapazitaet des jeweiligen Containers haben
; und dahinter die eingefuellten Stuecke.
; ((kapazitaet-1 stueck-a stueck-b ...) ...)
; Beispiel hier : ((120) (120) (100) ...)
 

;;; ===== ANFANG HILFSFUNKTIONEN =====
; ----- Hilfsfunktionen fuer bessere Lesbarkeit -----
; Die Restliste einer Liste nach dem Kopf ist natuerlich einfach nur (cdr ...)
(define (restliste liste) (cdr liste))
; Das erste Element entsprechend (car ...)
(define (erstes liste) (car liste))
; Die Restliste der Containerliste nach dem Kopf ist hier der inhalt
(define (inhalt container) (cdr container))
; Das erste Element ist hier die Kapazitaet
(define (kapazitaet container) (car container))
; Man fuellt ein Stueck ein mit (cons ... ...), hier aber hinter die Kapazitaet
; stueck x container -> container
(define (hinein stueck container) (cons (car container) (cons stueck (cdr container))))
; ----- Berechnung des gesamten Fuellung eines Containers -----
; Prinzip: Bilde eine Liste der Fuellungswerte und addiere diese.
(define (fuellung container) (apply + (inhalt container)))
; veraendert und ersetzt den veraenderten Container in der Container-Liste
(define
  (ersetze-container stueck container-liste index)
  (cond
    ((zero? index)
     (cons
      (hinein stueck (erstes container-liste))
      (restliste container-liste)))
    (else
     (cons
      (erstes container-liste)
      (ersetze-container stueck (restliste container-liste) (sub1 index))))))

; ----- Praedikate -----
; Fuer das Verstaendnis ist vermutlich einfacher, diese Prädikate auszugliedern:
; voll? prueft, ob alle Werte zusammen den Zielwert ergeben
(define (voll? container)    (= (kapazitaet container) (fuellung container)))
; zu-voll? prueft, ob alle Werte zusammen den Zielwert ueberschreiten
(define (zu-voll? container) (< (kapazitaet container) (fuellung container)))
; passt? prueft, ob alle Werte zusammen den Zielwert unterschreiten
(define (passt? container) (not (zu-voll? container)))
; ziel? prueft, ob bei allen containern alle Werte zusammen den Zielwert ergeben.
(define (ziel? container-liste)
  (if (null? container-liste) #t
      (and
       (voll? (car container-liste))
       (ziel? (restliste container-liste)))))
;;; ===== ENDE HILFSFUNKTIONEN =====

;;; ----- Funktionskopf -----
; 1. Schritt: Bestimmung des Funktionskopfes
(define
  (breitensuche stuecke container-liste)
  'wert-undefiniert)
; Testaufrufe:
(writeln "Schritt 1: ")
(breitensuche '(10 10) '((100 100) (120 120) (100 100)))
(breitensuche '(10 10) '((100 10) (120 10) (100 10)))

;;; ----- am Ziel? -----
; 2. Schritt: ich bin am Ziel, die Container sind exakt gefuellt.
; Hier ist also eine Bedingung zu überprüfen.
(define
  (breitensuche stuecke container-liste)
  (cond
   [(ziel? container-liste)    ; Bedingung 1
    container-liste]           ; ja-Fall: die container zurueckgeben
   [else
    'wert-undefiniert]                        ; der "sonst"-Fall
   )
  )

; Testaufrufe:
(writeln "Schritt 2: ")
(breitensuche '(10 10) '((100 100) (120 120) (100 100)))
(breitensuche '(10 10) '((100 10) (120 10) (100 10)))

;;; ----- am "Ende", aber nicht am Ziel -----
; 3. Schritt: es geht nicht weiter!
; Alle Stuecke wurden versucht, ohne dass sich die Container passend fuellen ließen.
(define
  (breitensuche stuecke container-liste)
  (cond
   [(ziel? container-liste)                   ; Bedingung 1
    container-liste]                          ; ja-Fall: die container zurueckgeben
   [(null? stuecke)                           ; 2.Bedingung
    #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
   [else
    'wert-undefiniert]                        ; der "sonst"-Fall
   )
  )

; Testaufrufe:
(writeln "Schritt 3: ")
(breitensuche '(10 10) '((100 100) (120 120) (100 100)))
(breitensuche '(10 10) '((100 10) (120 10) (100 10)))
(breitensuche '() '((100 10) (120 10) (100 10)))

;;; -----  -----
; 4. Schritt: es geht mit dem aktuellen Stueck weiter
; Nacheinander muss nun versucht werden, den Containern in allen Containerlisten
; das aktuelle Stueck hinzuzufügen. Dabei entstehen containeranzahl+1 neue mögliche
; container-listen, mit denen eine Warteschlange zu bilden ist.
; Die Steuerung uebernimmt der index der containerposition 0..n (n fuer gar nicht hinein)
; Also zwei neue Parameter: queue und index
; container-liste selbst kann dann aber entfallen = (car queue)
(define
  (breitensuche stuecke queue index)
  (cond
   [(ziel? (car queue))                       ; Bedingung 1
    (car queue)]                              ; ja-Fall: die container zurueckgeben
   [(null? stuecke)                           ; 2.Bedingung
    #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
   [(= index (length (car queue)))            ; alle versucht, nun auch ohne noch weiter
    (breitensuche                             ; und neues Element
   [(passt? (hinein (car stuecke) (list-ref (car queue) index)))
   [else
    'wert-undefiniert]                        ; der "sonst"-Fall
   )
  )

; Testaufrufe:
(writeln "Schritt 4: ")
(breitensuche '(10 10)  '(((100 100) (120 120) (100 100))) 0)
(breitensuche '(10 10) '(((100 10) (120 10) (100 10))) 0)
(breitensuche '() '(((100 10) (120 10) (100 10))) 0)



