;;;                                   "viele-rucksaecke-ts-entwicklung.scm"
;;; Demonstration der Entwicklung des Programmes zur Tiefensuche am Rucksackproblem
;;; hier fuer viele Rucksaecke, also muss eine Container-Liste verwendet werden.
(require (lib "breakpoint.ss"))
; ----- Liste der vorhandenen Stuecke -----
(define stuecke '(40 30 30 30 25 25 25 20 20 20 15 10 10 10 10 5))

; ----- die Container -----
; Da es nun nicht nur einen Container gibt, müssen wir eine Liste von Containern
; verwenden, die wiederum selbst aus Listen bestehen,
; die am Kopf die Kapazitaet des jeweiligen Containers haben
; und dahinter die eingefuellten Stuecke.
; ((kapazitaet-1 stueck-a stueck-b ...) ...)
; Beispiel hier : ((120) (120) (100) ...)
 

;;; ===== ANFANG HILFSFUNKTIONEN =====
; ----- Hilfsfunktionen fuer bessere Lesbarkeit -----
; Die Restliste einer Liste nach dem Kopf ist natuerlich einfach nur (cdr ...)
(define (restliste liste) (cdr liste))
; Das erste Element entsprechend (car ...)
(define (erstes liste) (car liste))
; Die Restliste der Containerliste nach dem Kopf ist hier der inhalt
(define (inhalt container) (cdr container))
; Das erste Element ist hier die Kapazitaet
(define (kapazitaet container) (car container))
; Man fuellt ein Stueck ein mit (cons ... ...), hier aber hinter die Kapazitaet
; stueck x container -> container
(define (hinein stueck container) (cons (car container) (cons stueck (cdr container))))
; ----- Berechnung des gesamten Fuellung eines Containers -----
; Prinzip: Bilde eine Liste der Fuellungswerte und addiere diese.
(define (fuellung container) (apply + (inhalt container)))
; veraendert und ersetzt den veraenderten Container in der Container-Liste
(define
  (ersetze-container stueck container-liste index)
  (cond
    ((zero? index)
     (cons
      (hinein stueck (erstes container-liste))
      (restliste container-liste)))
    (else
     (cons
      (erstes container-liste)
      (ersetze-container stueck (restliste container-liste) (sub1 index))))))

; ----- Praedikate -----
; Fuer das Verstaendnis ist vermutlich einfacher, diese Prädikate auszugliedern:
; voll? prueft, ob alle Werte zusammen den Zielwert ergeben
(define (voll? container)    (= (kapazitaet container) (fuellung container)))
; zu-voll? prueft, ob alle Werte zusammen den Zielwert ueberschreiten
(define (zu-voll? container) (< (kapazitaet container) (fuellung container)))
; passt? prueft, ob alle Werte zusammen den Zielwert unterschreiten
(define (passt? container) (not (zu-voll? container)))
; ziel? prueft, ob bei allen containern alle Werte zusammen den Zielwert ergeben.
(define (ziel? container-liste)
  (if (null? container-liste) #t
      (and
       (voll? (car container-liste))
       (ziel? (restliste container-liste)))))
;;; ===== ENDE HILFSFUNKTIONEN =====

;;; ----- Funktionskopf -----
; 1. Schritt: Bestimmung des Funktionskopfes
(define
  (tiefensuche stuecke container-liste)
  'wert-undefiniert)
; Testaufrufe:
(writeln "Schritt 1: ")
(tiefensuche '(10 10) '((100 100) (120 120) (100 100)))
(tiefensuche '(10 10) '((100 10) (120 10) (100 10)))

;;; ----- am Ziel? -----
; 2. Schritt: ich bin am Ziel, die Container sind exakt gefuellt.
; Hier ist also eine Bedingung zu überprüfen.
(define
  (tiefensuche stuecke container-liste)
  (cond
   [(ziel? container-liste)    ; Bedingung 1
    container-liste]           ; ja-Fall: die container zurueckgeben
   [else
    'wert-undefiniert]                        ; der "sonst"-Fall
   )
  )

; Testaufrufe:
(writeln "Schritt 2: ")
(tiefensuche '(10 10) '((100 100) (120 120) (100 100)))
(tiefensuche '(10 10) '((100 10) (120 10) (100 10)))

;;; ----- am "Ende", aber nicht am Ziel -----
; 3. Schritt: es geht nicht weiter!
; Alle Stuecke wurden versucht, ohne dass sich die Container passend fuellen ließen.
(define
  (tiefensuche stuecke container-liste)
  (cond
   [(ziel? container-liste)                   ; Bedingung 1
    container-liste]                          ; ja-Fall: die container zurueckgeben
   [(null? stuecke)                           ; 2.Bedingung
    #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
   [else
    'wert-undefiniert]                        ; der "sonst"-Fall
   )
  )

; Testaufrufe:
(writeln "Schritt 3: ")
(tiefensuche '(10 10) '((100 100) (120 120) (100 100)))
(tiefensuche '(10 10) '((100 10) (120 10) (100 10)))
(tiefensuche '() '((100 10) (120 10) (100 10)))

;;; ----- passt in den ersten (???) -----
; 4. Schritt: das aktuelle Stueck passt "in den ersten Container hinein" ist hier sinnlos.
; Da nicht auf die einzelnen Container über einen Namen zugegriffen wird, sondern
; auf die Container-Liste, muss waehrend der Bearbeitung klar sein, auf welche
; schon zugegriffen wurde und welcher der naechste ist, der noch zu versuchen ist.
; Dann wird weitergesucht mit diesem Stueck in dem besonderen Container.
; Wir beschreiben die noch zu versuchenden Container mit ihrem index in der Liste.
; wenn es keine gibt, muss noch versucht werden, ohne das erste Stueck weiter zu fuellen.
; Diese Bedingung muss vorher geprüft werden!
(define
  (tiefensuche stuecke container-liste index)
  (cond
   [(ziel? container-liste)                   ; Bedingung 1
    container-liste]                          ; ja-Fall: die container zurueckgeben
   [(null? stuecke)                           ; 2.Bedingung
    #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
   [(= index (length container-liste))        ; 3.Bedingung: 
    (tiefensuche                              ; ohne das stueck versuchen.
     (restliste stuecke)
     container-liste
     alternativen)]
   [(if                                      ; 4.Bedingung: passt in den aktuellen
     (passt? (hinein (erstes stuecke) (list-ref container-liste index)))
     (tiefensuche
      (cdr stuecke)      
      (ersetze-container
       (erstes stuecke)
       container-liste
       index)
      0)                                      ; nun alle Container wieder versuchen
     #f)]
   [else                                      ; und sonst?
    'wert-undefiniert]                        ; der "sonst"-Fall
    )
  )

; Testaufrufe:
(writeln "Schritt 4: ")
(tiefensuche stuecke '((100 10) (120 10) (100 10)) 0)

;;; ----- sonst? -----
; 5. Schritt: das aktuelle Stueck passt nicht in diesen.
(define
  (tiefensuche stuecke container-liste index)
  (cond
   [(ziel? container-liste)                   ; Bedingung 1
    container-liste]                          ; ja-Fall: die container zurueckgeben
   [(null? stuecke)                           ; 2.Bedingung
    #f]                                       ; ja-Fall. #f kennzeichnet Misserfolg.
   [(= index (length container-liste))        ; 3.Bedingung: 
    (tiefensuche                              ; ohne das stueck versuchen.
     (restliste stuecke)
     container-liste
     0)]
   [(if                                      ; 4.Bedingung: passt in den aktuellen
     (passt? (hinein (erstes stuecke) (list-ref container-liste index)))
     (tiefensuche
      (cdr stuecke)      
      (ersetze-container
       (erstes stuecke)
       container-liste
       index)
      0)                                      ; nun alle Container wieder versuchen
     #f)]
    [else                                      ; und sonst?
     (tiefensuche                              ; einfach beim naechsten weiter versuchen.
      stuecke
      container-liste
      (add1 index))]
    )
  )

; Testaufrufe:
(writeln "Schritt 5: ")
(tiefensuche stuecke '((100) (120)) 0)
(tiefensuche stuecke '((100) (120) (100)) 0)
;(tiefensuche stuecke '((100) (120) (120)) 0)

