;;;                                   "zwei-rucksaecke-ts-entwicklung.scm"
;;; Demonstration der Entwicklung des Programmes zur Tiefensuche am Rucksackproblem
;;; hier fuer zwei Rucksaecke container-1 und container-2
(require (lib "breakpoint.ss"))
; ----- Liste der vorhandenen Stuecke -----
(define stuecke '(40 30 30 30 25 25 25 20 20 20 15 10 10 10 5))

; ----- die Container -----
; Da es nun nicht nur einen gibt, müssen sie aus Listen bestehen,
; die am Kopf die Kapazitaet haben des jeweiligen Containers haben
; und dahinter die eingefuellten Stuecke.
; (kapazitaet-1 stueck-a stueck-b ...)
; Beispiel hier : (120) (100)
 

;;; ===== ANFANG HILFSFUNKTIONEN =====
; ----- Hilfsfunktionen fuer bessere Lesbarkeit -----
; Die Restliste einer Liste nach dem Kopf ist natuerlich einfach nur (cdr ...)
(define (restliste liste) (cdr liste))
; Das erste Element entsprechend (car ...)
(define (erstes liste) (car liste))
; Die Restliste der Containerliste nach dem Kopf ist hier der inhalt
(define (inhalt container) (cdr container))
; Das erste Element ist hier die Kapazitaet
(define (kapazitaet container) (car container))
; Man fuellt ein Stueck ein mit (cons ... ...), hier aber hinter die Kapazitaet
; stueck x container -> container
(define (hinein stueck container) (cons (car container) (cons stueck (cdr container))))
; ----- Berechnung des gesamten Fuellung eines Containers -----
; Prinzip: Bilde eine Liste der Fuellungswerte und addiere diese.
(define (fuellung container) (apply + (inhalt container)))

; ----- Praedikate -----
; Fuer das Verstaendnis ist vermutlich einfacher, diese Prädikate auszugliedern:
; ziel? prueft, ob alle Werte zusammen den Zielwert ergeben
(define (ziel? container)    (= (kapazitaet container) (fuellung container)))
; zu-voll? prueft, ob alle Werte zusammen den Zielwert ueberschreiten
(define (zu-voll? container) (< (kapazitaet container) (fuellung container)))
; passt? prueft, ob alle Werte zusammen den Zielwert unterschreiten
(define (passt? container) (not (zu-voll? container)))


;;; ===== ENDE HILFSFUNKTIONEN =====

;;; ----- Funktionskopf -----
; 1. Schritt: Bestimmung des Funktionskopfes
(define
  (tiefensuche stuecke container-1 container-2)
  'wert-undefiniert)
; Testaufrufe:
(writeln "Schritt 1: ")

;;; ----- am Ziel? -----
; 2. Schritt: ich bin am Ziel, die Container sind exakt gefuellt.
; Hier ist also eine Bedingung zu überprüfen. --> if
(define
  (tiefensuche stuecke container-1 container-2)
  (cond
   [(and
     (ziel? container-1)
     (ziel? container-2))                     ; Bedingung 1
    (list container-1 container-2)]           ; ja-Fall: die container zurueckgeben
   [else
    'wert-undefiniert]                        ; der "sonst"-Fall
   )
  )

; Testaufrufe:
(writeln "Schritt 2: ")

;;; ----- am "Ende", aber nicht am Ziel -----
; 3. Schritt: es geht nicht weiter!
; Alle Stuecke wurden versucht, ohne dass sich die Container passend fuellen ließen.
(define
  (tiefensuche stuecke container-1 container-2)
  (cond
   [(and
     (ziel? container-1)
     (ziel? container-2))                     ; Bedingung 1
    (list container-1 container-2)]           ; ja-Fall: die container zurueckgeben
    [(null? stuecke)                          ; 2.Bedingung
     #f]                                      ; ja-Fall. #f kennzeichnet Misserfolg.
   [else
    'wert-undefiniert]                        ; der "sonst"-Fall
   )
  )

; Testaufrufe:
(writeln "Schritt 3: ")

;;; ----- passt in den ersten -----
; 4. Schritt: das aktuelle Stueck passt in den ersten Container hinein.
; Es wird weitergesucht mit diesem Stueck im Container.
; Da aber Misserfolgmöglich ist, muss der Aufruf als Bedingung eingesetzt sein.
(define
  (tiefensuche stuecke container-1 container-2)
  (cond
   [(and
     (ziel? container-1)
     (ziel? container-2))                     ; Bedingung 1
    (list container-1 container-2)]           ; ja-Fall: die container zurueckgeben
    [(null? stuecke)                          ; 2.Bedingung
     #f]                                      ; ja-Fall. #f kennzeichnet Misserfolg.
    [(if                                      ; 3.Bedingung: passt in den ersten
      (passt? (hinein (erstes stuecke) container-1))
      (tiefensuche
       (cdr stuecke)
       (hinein (erstes stuecke) container-1)
       container-2)
      #f)]
    [else                                      ; und sonst?
     'wert-undefiniert]                        ; der "sonst"-Fall
    )
  )

; Testaufrufe:
(writeln "Schritt 4: ")

;;; ----- passt in den zweiten -----
; 5. Schritt: das aktuelle Stueck passt in den zweiten Container hinein.
; Es wird weitergesucht mit diesem Stueck im Container.
; Da aber Misserfolgmöglich ist, muss der Aufruf als Bedingung eingesetzt sein.
(define
  (tiefensuche stuecke container-1 container-2)
  (cond
   [(and
     (ziel? container-1)
     (ziel? container-2))                     ; Bedingung 1
    (list container-1 container-2)]           ; ja-Fall: die container zurueckgeben
    [(null? stuecke)                          ; 2.Bedingung
     #f]                                      ; ja-Fall. #f kennzeichnet Misserfolg.
    [(if                                      ; 3.Bedingung: passt in den ersten
      (passt? (hinein (erstes stuecke) container-1))
      (tiefensuche
       (cdr stuecke)
       (hinein (erstes stuecke) container-1)
       container-2)
      #f)]
    [(if                                      ; 4.Bedingung: passt in den zweiten
      (passt? (hinein (erstes stuecke) container-2))
      (tiefensuche
       (cdr stuecke)
       container-1
       (hinein (erstes stuecke) container-2))
      #f)]
    [else                                      ; und sonst?
     'wert-undefiniert]                        ; der "sonst"-Fall
    )
  )

; Testaufrufe:
(writeln "Schritt 4: ")

;;; ----- der sonst-Fall -----
; 6. Schritt: das Stueck passte in keinen Container.
(define
  (tiefensuche stuecke container-1 container-2)
  (cond
   [(and
     (ziel? container-1)
     (ziel? container-2))                     ; Bedingung 1
    (list container-1 container-2)]           ; ja-Fall: die container zurueckgeben
    [(null? stuecke)                          ; 2.Bedingung
     #f]                                      ; ja-Fall. #f kennzeichnet Misserfolg.
    [(if                                      ; 3.Bedingung: passt in den ersten
      (passt? (hinein (erstes stuecke) container-1))
      (tiefensuche
       (cdr stuecke)
       (hinein (erstes stuecke) container-1)
       container-2)
      #f)]
    [(if                                       ; 4.Bedingung: passt in den zweiten
      (passt? (hinein (erstes stuecke) container-2))
      (tiefensuche
       (cdr stuecke)
       container-1
       (hinein (erstes stuecke) container-2))
      #f)]
    [else                                      ; und sonst?
     (tiefensuche                              ; weiter suchen ...
      (restliste stuecke)                      ; ... ohne das erste Element ...
      container-1                              ; ... mit der bisherigen Fuellung
      container-2)]                            ; ... und dem selben Zielwert.
    )
  )

; Testaufrufe:
(writeln "Schritt 6: ")
(tiefensuche stuecke '(40) '(30))
(tiefensuche stuecke '(120) '(100))
(tiefensuche stuecke '(220) '(180))

