;;;                                    <clp-mit-Tiefensuche.scm>
;;; Es wird das Container - Lade - Problem untersucht.
;;; Der Suchraum wird nicht real angelegt, sondern jeder Knoten wird
;;; genau dann erzeugt, wenn er gebraucht wird.
;;; Knoten sind die ja/nein - Verzweigungen fr hinein oder nicht.
(require (lib "breakpoint.ss"))
;;;
;;; Dabei wird zwar prinzipiell die einfache Lsung des
;;; Rucksackproblems mit Tiefensuche verwendet. Allerdings haben
;;; wir nun ein Optimierungsproblem, so dass einerseits eine
;;; vollstndige Suche notwendig ist und wir zudem eine Bewertungs-
;;; funktion brauchen. Hierfr gbe es je nach Anwendungstyp zwei
;;; verschiedene Lsungen: Einmal die Zahl und Gre der bentigten
;;; Container oder umgekehrt die Summe der Gewichte (Gren) der
;;; eingeladenen Teile.
;;; 
;;; Bewertungsfunktion:
;;; Summe der Gewichte, wird nur bei gefllten Containern aufgerufen
;;;
;;; Datenstruktur
;;; Da die Container nicht einer nach dem anderen befllt werden,
;;; muss  man sich zu jedem den Fllstand merken. Daher ist eine
;;; Assoziationsliste mit Paaren aus Groesse und Fllteilen angemessen.
;;; Container: '((100 (40 30 10)) (100 (40 20)))
(define Container '((100 (40 30 10)) (100 (40 20))))

;;; ===== Fuellung =====
; Ein Container --> seine Fllung
(define
  (Fuellung ein-Container)
  (cond
    ((null? (cadr ein-Container)) 0)
    (else
     (apply + (cadr ein-Container)))))

;;; ===== Bewertung =====
; Container --> Summe Fllteile
(define
  (Bewertung alle-Container)
  (cond
    ((null? alle-Container) 0)
    (else
     (apply + (map Fuellung alle-Container)))))

; Test
;(Bewertung Container)

;;; ===== Wege ? =====
; Das Problem ist, die Wege im Graphen richtig zu beschreiben.
; Ein Weg ist der Aufbau einer zulssigen Belegung.
; Alternativen sind: Teil nicht einbauen, Teil im ersten einbauen,
; im zweiten ... usw. Die Funktion versuche muss diese Flle zurck
; liefern.

;;; ===== versuche =====
; Es wird versucht, das Teil in einen der Container einzufgen. 
; Dabei werden alle Mglichkeiten zurckgegeben.(-->Alternativen)
(define
  (versuche Teil Container)
  (let
      versuche                             ; lokal verdeckt global
    ((Container Container)
     (bearbeitet '())
     (akku (list Container)))                     ; die waren zulssig
    (cond
      ((null? Container)
       ; fertig.
       akku)
      ((> (+ Teil (Fuellung (car Container))) (caar Container))
       ; passt nicht hinein, nchsten versuchen.
       (versuche (cdr Container) (cons (car Container) bearbeitet) akku))
      (else
       ; passt hinein, diese Kombination muss im akku gespeichert werden.
       (versuche
        (cdr Container)
        (cons (car Container) bearbeitet)
        (cons
         (append
          (reverse bearbeitet)
          (cons
           (cons
            (caar Container)
            (list
             (cons
              Teil
              (cadar Container))))
           (cdr Container)))
         akku))))))

; Test:
;(versuche 80 '((100 (40 30 10)) (100 (40 20))))
;(versuche 25 '((100 (40 30 10)) (100 (40 20))))
;(versuche 10 '((100 (40 30 10)) (100 (40 20))))
;;; ===== Tiefensuche =====
(define
  (Container-packen Container Teile)
  (let t-s
    ((Alternativen (versuche (car Teile) Container))
     (Teile (cdr Teile)))
    (cond
      ((null? Teile)
       (Ausgabe Alternativen)
       #f)
      ((null? Alternativen) #f)
      (else
       ; Schritt in die Tiefe mit Zwang zur Alternative
       ;(bkpt "Tiefe" (versuche (car Teile) (car Alternativen)) (cdr Teile))
       (t-s (versuche (car Teile) (car Alternativen)) (cdr Teile))
       ;(bkpt "Alternative" (cdr Alternativen) Teile)
       (t-s (cdr Alternativen) Teile)))))

;;; ===== Ausgabe =====
; Hier werden nun die besten verwaltet. "bel": Eine globale Variable!
(define beste-Bewertung 0)
(define beste-Alternative #f)
(define 
  (Ausgabe Alternativen)
  (cond
    ((null? Alternativen) (void))
    ((> (Bewertung (car Alternativen)) beste-Bewertung)
     (writeln beste-Bewertung " bei " beste-Alternative)
     (set! beste-Bewertung (Bewertung (car Alternativen)))
     (set! beste-Alternative (car Alternativen))
     (Ausgabe (cdr Alternativen)))
    (else
     (Ausgabe (cdr Alternativen)))))
  

(define Container '((100 ()) (100 ()) (120 ()) (120 ())))
(define Teile '(70 30 20 30 50 10 60 20 30 50 30 40 10 50 60 90))
(Container-packen Container Teile)
beste-Bewertung
beste-Alternative