;;; "fuelle-viele-bs.scm"
;;;  Hier also mit Breitensuche !!!
; ein Container ist eine Liste, die zun?chst sein Fassungsverm?gen
; enth?lt. Werden St?cke eingef?gt, dann geschieht das hinter dieser
; ersten Zahl der Liste. Beispiel:
; (100 20 30 30) ist ein Container mit dem Fassungsverm?gen 100, der 
; zwei 30-er - St?cke und ein 20-er - St?ck enth?lt.

(require (lib "breakpoint.ss"))

;;; ----- ist-voll? -----
; untersucht, ob ein Container voll ist.
(define
  (ist-voll? container)
  (= (car container) (apply + (cdr container))))

;;; ----- zu-voll? -----
; untersucht, ob ein Container zu (!) voll ist.
(define
  (zu-voll? container)
  (< (car container) (apply + (cdr container))))

;;; ----- alle-voll? -----
; pr?ft, ob alle Container voll sind.
(define
  (alle-voll? alle-container)
  (cond
    ((null? alle-container) #t)
    ((ist-voll? (car alle-container))
     (alle-voll? (cdr alle-container)))
    (else #f)))

;;; ----- fuelle -----
; fuellt das stueck in den Container. Es ist zu beachten, dass
; vorn in der Containerliste das Fassungsverm?gen steht.
(define
  (fuelle stueck container)
  (cons
   (car container)
   (cons
    stueck
    (cdr container))))

;;; ----- bearbeite -----
; erzeugt die Containerliste, bei der das Stueck in den aktuell zu
; bearbeitenden Container gef?llt wird.
; Da der vordere Teil wegen der vorherigen Bearbeitung verkehrt herum
; ist, wird vorn vorher gedreht.
(define
  (bearbeite stueck vorn hinten)
  (append
   (reverse vorn)
   (cons
    (fuelle stueck (car hinten))
    (cdr hinten))))

;(bearbeite 20 '((300) (200) (100)) '((50) (60)))
;;; ----- generiere-nachfolger -----
; erzeugt die Liste aller nachfolgenden m?glichen Belegungen.
; Dabei werden ?berbelegungen nicht generiert. Es muss auch die
; M?glichkeit generiert werden, bei der das Stueck nicht verwendet
; wird.
; Das Problem ist: Es darf immer nur einer der Container bef?llt werden!
; Daher <vorn> <fuelle erstes von hinten> <restliche hinten>
(define
  (generiere-nachfolger stueck vorn hinten)
  (cond
    ((null? hinten) ; noch der Fall "nicht eingebaut";
     (list (reverse vorn)))
    ((zu-voll? (fuelle stueck (car hinten))) ; weglassen
     (generiere-nachfolger
      stueck
      (cons (car hinten) vorn)
      (cdr hinten)))
    (else
     (cons
      (bearbeite stueck vorn hinten)
      (generiere-nachfolger
       stueck
       (cons (car hinten) vorn)
       (cdr hinten))))))

;;; ----- die Suchfunktion -----
(define
  (suche alle-stuecke alternativen akku)
  ;(bkpt  alle-stuecke alternativen akku)
  (cond 
    ((null? alternativen) 
     (suche (cdr alle-stuecke) akku '()))
    ((alle-voll? (car alternativen)) (car alternativen))
    ((null? alle-stuecke) #f)
    (else
     (suche
      alle-stuecke
      (cdr alternativen)
      (append
       (generiere-nachfolger
        (car alle-stuecke)
        '()
        (car alternativen))
       akku)))))


;;; ----- Testaufrufe -----
(suche
 '(30 30 30 20 20 20 20 10 10 10 5 5 5 5)
 '(((100) (100)))
 '())

;(suche
; '(30 30 30 30 30 30 30 30 30 30 30 30 30 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 5 5 5 5 5 5 5 5)
; '(((100) (100) (100) (100) (100) (100) (100) (100) (100) (100) (100) (100)))
; '())

;(suche
; '(90 90 90 80 80 70 70 70 70 30 30 30 30 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 5 5 5 5 5 5 5 5)
; '(((100) (100) (100) (100) (100) (100) (100) (100) (100) (100) (100) (100)))
; '())

