;;; "clp-ga-selektion.scm"
;;; ===== Selektion =====
; Die Individuen im Pool werden so einer Selektion unterworfen, dass
; zunchst die besten (beste-%) ausgewhlt werden und anschliessend
; der Pool wieder auf die Anzahl aufgefllt wird, indem
; zufllige Paare gegeneinander antreten.
(define
  (Selektion pool anzahl beste-%)
  (let
      ((anzahl-beste (inexact->exact (round (* beste-% anzahl)))))
    (append
     (bestimme-beste anzahl-beste pool '() '())
     (antreten (- anzahl anzahl-beste) pool '()))))
;;; ----- (bestimme-beste -----
; bestimmt die anzahl besten im Pool
; die Funktion wird anzahl oft durchlaufen und legt jeweils
; die besten vorn in ablage ab.
(define
  (bestimme-beste anzahl liste akku ablage)
  (cond
    ((zero? anzahl)
     ; alle Durchlufe sind fertig, Ergebnis ist in akku
     akku)
    ((null? ablage)
     ; erster Durchlauf, erstes in ablage
     (bestimme-beste anzahl (cdr liste) akku (cons (car liste) ablage)))
    ((null? liste)
     ; ein Durchlauf fertig, das erste aus ablage ist das aktuelle beste
     (bestimme-beste (sub1 anzahl) (cdr ablage) (cons (car ablage) akku) '()))
    ((> (bewerte (car liste)) (bewerte (car ablage)))
     ; das aktuelle ist besser und kann nach vorn in die ablage
     (bestimme-beste
      anzahl
      (cdr liste)
      akku
      (cons (car liste) ablage)))
    (else
     ; das aktuelle ist nicht besser, also muss es hinter das erste
     (bestimme-beste
      anzahl
      (cdr liste)
      akku
      (cons (car ablage) (cons (car liste) (cdr ablage)))))))
;;; ----- antreten -----
; lsst anzahl oft zufllig ausgewhlte Individuen gegeneinander
; antreten
(define
  (antreten anzahl liste akku)
  (let
      ((erstes (list-ref liste (random (length liste))))
       (zweites (list-ref liste (random (length liste)))))
    (cond
      ((zero? anzahl) akku)
      ((> (bewerte erstes) (bewerte zweites))
       (antreten (sub1 anzahl) liste (cons erstes akku)))
      (else
       (antreten (sub1 anzahl) liste (cons zweites akku))))))

