;;; "clp-ga.scm"
; Neue Modellierung fr das CLP mit Genetischen Algorithmen
; Vorgegeben ist eine Liste mit Containern und eine Liste mit Stcken
; berschreitungen werden akzeptiert, allerdings mit mehrfachem Gewicht
; als malus bewertet.
; Die nicht eingebauten Stcke liegen dazu in einem "Container" 
; mit dem Fassungsvermgen 0.

;;; Die vorliegende Lsung weicht dadurch von der Standardlsung ab, dass
;;; keine Individuen untereinander "Gene" austauschen, sondern nur
;;; die Container eines (!) Individuums jeweils. Dadurch erspart man sich
;;; die Betrachtungen der Zulssigkeit. Der Pool ermglicht es aber, schnell
;;; ungnstige Vorgaben auszuschlieen.
(require (lib "breakpoint.ss"))

(define
  *Stuecke*
  '(10 20 30 40 10 20 30 40 50 80 10 20 30 50
       80 10 20 30 40 10 20 30 40 50 80 10 20 30 40 10 20 30 40 50 80))

(define
  *Container*
  (cons
   (cons 0 *Stuecke*)
   '((100)
     (100)
     (100)
     (160)
     (160))))

;;; ===== Bewertung =====
; Die Elemente des POOLS werden so bewertet, dass vom Wert
; der eingefllten Stcke der Wert der berschrittenen mehrfach (faktor) abgezogen wird.
(define
  (bewerte container-liste)
  (cond
    ((null? container-liste) 0)
    ((zero? (caar container-liste)) ; restliche Stuecke rausschmeien
     (bewerte (cdr container-liste)))
    ((< (caar container-liste) (apply + (cdar container-liste))) ; zu voll!
     (+ (bewerte (cdr container-liste))
        (apply + (cdar container-liste))
        (* 3 (- (caar container-liste) (apply + (cdar container-liste))))))
    (else
     (+ (bewerte (cdr container-liste))
        (apply + (cdar container-liste))))))

;(bewerte '((0 30 20 10 40 30 20 10 80 50 30 20 10 80 50 40 30 20 10 40 30 20 10)
; (100 10 50)
; (100 40 80)
; (100 20 30 40)
; (160 10 20 30 40)
; (160 50 80)))
;;; ===== Mutation =====
; Bei der Mutation wird ein zufllig gewhltes Stck
; aus einem zufllig gewhlten Container entnommen und
; in einen zufllig gewhlten anderen getan.
(define
  (Mutation pool akku)
  (cond
    ((null? pool) akku)
    (else
     (Mutation
      (cdr pool)
      (cons
        (mutiere
         (random (length (car pool))) 
         (random (length (car pool)))
         (car pool))
        (cons
         (car pool)
         akku))))))

;;; ----- mutiere -----
; Fhrt die Mutation durch
; zahl x zahl x liste --> liste
(define
  (mutiere von-container-nr nach-container-nr container-liste)
  (cond
    ((= von-container-nr nach-container-nr)
     container-liste) ; sinnlos!
    ((leer? von-container-nr container-liste)
     container-liste) ; geht nicht!
    (else
     (letrec
         ((container (list-ref container-liste von-container-nr))
          (stueck-nr (add1 (random (sub1 (length container)))))
          (stueck (list-ref container stueck-nr)))
       (entferne
        stueck
        von-container-nr
        (fuelle
         stueck
         nach-container-nr
         container-liste))))))

;;; ----- leer? -----
; prft, ob der container mit der bergebenen nummer leer ist.
; nr: 0 .. n-1
(define
  (leer? nr container-liste)
  (null? (cdr (list-ref container-liste nr))))

;;; ----- entferne -----
; Wird ber das remove hinaus gebraucht, um zu verhindern,
; dass das erste Element -die Containergre- entfernt wird.
(define
  (entferne stueck nummer container-liste)
  (cond
    ((null? container-liste) (error "falscher index!"))
    ((negative? nummer) (error "falscher index!"))
    ((zero? nummer)
     (cons
      (cons
       (caar container-liste) ; Fassungsvermgen
       (remove stueck (cdar container-liste))) ; Inhalte
      (cdr container-liste)))
    (else
     (cons
      (car container-liste)
      (entferne stueck (sub1 nummer) (cdr container-liste))))))

;;; ----- fuelle -----
; Wird ber das cons hinaus gebraucht, um zu verhindern, dass das 
; stueck als erstes Element -die Containergre- verwendet wird.
(define
  (fuelle stueck nummer container-liste)
  (cond
    ((null? container-liste) (error "falscher index!"))
    ((negative? nummer) (error "falscher index!"))
    ((zero? nummer)
     (cons
      (cons
       (caar container-liste) ; Fassungsvermgen
       (cons stueck (cdar container-liste))) ; Inhalte
      (cdr container-liste)))
    (else
     (cons
      (car container-liste)
      (fuelle stueck (sub1 nummer) (cdr container-liste))))))

;;; ===== POOL =====
; Erzeugen des POOLS.
; Es werden n verschiedene zufllig gefllte Containerlisten erzeugt.
(define
  (erzeuge-pool container-liste n)
  (cond
    ((positive? n)
     (cons
      (erzeuge-pool-element
       (umbau ; verndert die stueckeliste
        (cdar container-liste)
        (random (length (cdar container-liste)))
        '()) ; da endrekursiv !
       (cdr container-liste)
       '()) ; da endrekursiv !
      (erzeuge-pool container-liste (sub1 n))))
    (else  '())))

;;; ----- umbau -----
; die Liste der vorhandenen Stcke wird an der bergebenen stelle auseinander
; geschnitten und wieder zusammengesetzt. Reihenfolge z.T. gendert.
(define
  (umbau stuecke stelle akku)
  (cond
    ((null? stuecke) akku)
    ((zero? stelle)
     (append stuecke akku))
    (else
     (umbau (cdr stuecke) (sub1 stelle) (cons (car stuecke) akku)))))

;;; ----- erzeuge-pool-element -----
; Aus der Stckeliste werden in jeden Container gerade soviel Stcke getan
; bis er nicht berfllt ist.
; Achtung: die Containerliste enthlt hier nur die tatschlichen Container! 
(define
  (erzeuge-pool-element stuecke-liste container-liste akku)
  (cond
    ((null? container-liste) ; fertig, noch restliche Stcke davor!
     (cons
      (cons 0 stuecke-liste)
      (reverse akku)))
    ((null? stuecke-liste) ; sollte nicht sein, keine Stcke mehr!
     (cons
      (cons 0 stuecke-liste)
      (reverse akku)))
    ((>                     ; zu viel?
      (apply + (cons (car stuecke-liste) (cdar container-liste)))
      (caar container-liste))
     (erzeuge-pool-element
      stuecke-liste
      (cdr container-liste)
      (cons (car container-liste) akku)))
    (else                   ; rein das Stck!
     (erzeuge-pool-element
      (cdr stuecke-liste)
      (cons
       (cons
        (caar container-liste) ; Fassung
        (cons
         (car stuecke-liste)
         (cdar container-liste)))  ; mit Stck
       (cdr container-liste))
      akku))))
;;; ----- Test -----
;(erzeuge-pool *Container* 1)
;(Test-reine-Mutation (car (erzeuge-pool *Container* 1)) *Container* 100)

;;; ===== Selektion =====
; Die Individuen im Pool werden so einer Selektion unterworfen, dass
; zunchst die besten (beste-%) ausgewhlt werden und anschliessend
; der Pool wieder auf die Anzahl aufgefllt wird, indem
; zufllige Paare gegeneinander antreten.
(define
  (Selektion pool anzahl beste-%)
  (let
      ((anzahl-beste (inexact->exact (round (* beste-% anzahl)))))
    (append
     (bestimme-beste anzahl-beste pool '() '())
     (antreten (- anzahl anzahl-beste) pool '()))))
;;; ----- (bestimme-beste -----
; bestimmt die anzahl besten im Pool
; die Funktion wird anzahl oft durchlaufen und legt jeweils
; die besten vorn in ablage ab.
(define
  (bestimme-beste anzahl liste akku ablage)
  (cond
    ((zero? anzahl)
     ; alle Durchlufe sind fertig, Ergebnis ist in akku
     akku)
    ((null? ablage)
     ; erster Durchlauf, erstes in ablage
     (bestimme-beste anzahl (cdr liste) akku (cons (car liste) ablage)))
    ((null? liste)
     ; ein Durchlauf fertig, das erste aus ablage ist das aktuelle beste
     (bestimme-beste (sub1 anzahl) (cdr ablage) (cons (car ablage) akku) '()))
    ((> (bewerte (car liste)) (bewerte (car ablage)))
     ; das aktuelle ist besser und kann nach vorn in die ablage
     (bestimme-beste
      anzahl
      (cdr liste)
      akku
      (cons (car liste) ablage)))
    (else
     ; das aktuelle ist nicht besser, also muss es hinter das erste
     (bestimme-beste
      anzahl
      (cdr liste)
      akku
      (cons (car ablage) (cons (car liste) (cdr ablage)))))))
;;; ----- antreten -----
; lsst anzahl oft zufllig ausgewhlte Individuen gegeneinander
; antreten
(define
  (antreten anzahl liste akku)
  (let
      ((erstes (list-ref liste (random (length liste))))
       (zweites (list-ref liste (random (length liste)))))
    (cond
      ((zero? anzahl) akku)
      ((> (bewerte erstes) (bewerte zweites))
       (antreten (sub1 anzahl) liste (cons erstes akku)))
      (else
       (antreten (sub1 anzahl) liste (cons zweites akku))))))

;;; ===== Generationen-folge =====
(define
  (Generationen-folge pool anzahl)
  (cond
    ((zero? anzahl)
     pool)
    (else
     (Generationen-folge (Generation pool) (sub1 anzahl)))))

;;; ----- generation -----
; Fhrt einen Generationenschritt aus.
; es werden alle einmal mutiert und die besten aus den vorher vorhandenen und 
; den mutierten ausgewhlt.
(define
  (Generation pool)
  (Selektion
   (Mutation pool '())
   (length pool)
   0.30))


;;; ===== Test =====
; 620 ist maximum!
(let
    ((beste (car (bestimme-beste 1 (Generationen-folge (erzeuge-pool *Container* 40) 20) '() '()))))
  (display beste) (newline)
  (display (bewerte beste)) (newline))
