;;;                                      <GA Money Funktionen.scm>
(require (lib "breakpoint.ss"))
; Das Setzen des Zufallszahlengenerators sollte im HP erfolgen.
;;; ===== erzeuge-Individuum ========================================
; Erzeugt ein Individuum der Population.
;;; Alle Individuen bei SEND MORE MONEY sind beliebige Belegungen der
;;; Buchstaben mit Ziffern.
;;; Daraus wird eine Assoziationsliste gebildet.

(define Buchstabenliste '(s e n d m o r y))

(define
  (erzeuge-Individuum)
  (let
      loop
    ((liste Buchstabenliste)
     (akku ()))
    (cond
      ((null? liste)
       akku)
      (else
       (loop
        (cdr liste)
        (cons
         (list (car liste) (random 10))
         akku))))))

;;; ===== bewerte ===================================================
; Bewertet ein Individuum mit einer Bewertungsziffer.
; Liefert eine Liste aus Individuum und Bewertungszahl zurck
; Zunchst wird fr jede Mehrfachbelegung ein Wert hinzugezhlt, 
; dann fr jeden Fehler bei einer Berechnung ein anderer Wert,
; ausserdem fr vorn stehende Nullen ein dritter Wert.

(define
  (bewerte Individuum)
  (list
   Individuum
   (+ 
    (Anzahl-doppelte Individuum)
    (Anzahl-Fehler Individuum)
    (vorn-null Individuum))))

(define
  (vorn-null Individuum)
  (+
   (if (zero? (cadr (assoc 'm Individuum))) 5 0)
   (if (zero? (cadr (assoc 's Individuum))) 5 0)))
   
(define
  (Anzahl-doppelte Belegung)
  (define
    (tief-member paar paar-liste)
    (cond
      ((null? paar-liste) #f)
      ((equal? (cadr paar) (cadr (car paar-liste))) #t)
      (else
       (tief-member paar (cdr paar-liste)))))
  (let
      loop
    ((Belegung Belegung)
     (doppelte 1)
     (akku 0))
    (cond
      ((null? (cdr Belegung))
       akku)
      (else
       (loop
        (cdr Belegung)
        doppelte
        (if (tief-member (car Belegung) (cdr Belegung))
            (+ doppelte akku)
            akku))))))

(define
  (Wort->Zahl Wort Belegung akku)
  (cond
    ((null? Wort) akku)
    (else
     (Wort->Zahl 
      (cdr Wort) 
      Belegung
      (+ (* 10 akku) (cadr (assoc (car Wort) Belegung)))))))

(define
  (verschieden zahl-1 zahl-2)
  (cond
    ((and (zero? zahl-1) (zero? zahl-2)) 0)
    ((= (modulo zahl-1 10) (modulo zahl-2 10))
     (verschieden (quotient zahl-1 10) (quotient zahl-2 10)))
    (else
     (add1 (verschieden (quotient zahl-1 10) (quotient zahl-2 10))))))

(define
  (Anzahl-Fehler Belegung)
  (let                         ;'(s e n d m o r y)
      ((Ergebnis
        (+
         (Wort->Zahl '(s e n d) Belegung 0)
         (Wort->Zahl '(m o r e) Belegung 0)))
       (Ziel (Wort->Zahl '(m o n e y) Belegung 0)))
    (verschieden Ergebnis Ziel)))

;;; ===== Darstellung ===============================================
; Stellt eine Belegung mit Berechnung dar.
(define
  (Darstellung Belegung-mit-Bewertung)
  (writeln Belegung-mit-Bewertung)
  (writeln (Wort->Zahl '(s e n d) (car Belegung-mit-Bewertung) 0))
  (writeln (Wort->Zahl '(m o r e) (car Belegung-mit-Bewertung) 0))
  (writeln (+
            (Wort->Zahl '(s e n d) (car Belegung-mit-Bewertung) 0)
            (Wort->Zahl '(m o r e) (car Belegung-mit-Bewertung) 0)))
  (writeln)
  (writeln (Wort->Zahl '(m o n e y) (car Belegung-mit-Bewertung) 0))
  (writeln "=============================")
  )
;;; ===== mutiere ===================================================
; Mutiert ein einzelnes Individuum.
; Dazu wird in der Belegung zufllig ein zufllig gewhlter einzelner
; Wert gendert.
; Individuum -> Individuum
(define
  (mutiere Individuum)
  (let loop
    ((Individuum Individuum)
     (Position (random (length Individuum))))
    (cond
      ((zero? Position)
       (cons (list (caar Individuum) (random 10)) (cdr Individuum)))
      (else 
       (cons
        (car Individuum)
        (loop (cdr Individuum) (sub1 Position)))))))
  
;;; ===== kreuze ====================================================
; Kreuzt zwei einzelne Individuen.
; Individuum x Individuum -> (list Individuum Individuum)
(define
  (kreuze Individuum-1 Individuum-2)
  (let ((Laenge (length Individuum-1)))   
    (let loop
      ((ab-Position (random Laenge))
       (bis-Position (random Laenge))
       (Individuum-1 Individuum-1)
       (Individuum-2 Individuum-2)
       (akku-1 ())
       (akku-2 ()))
      (cond
        ((> ab-Position bis-Position) ; Reihenfolge-Korrektur
         (loop bis-Position ab-Position Individuum-1 Individuum-2 akku-1 akku-2))
        ((null? Individuum-1)
         (list (reverse akku-1) (reverse akku-2)))
        ((positive? ab-Position)             ;weiter durchgehen
         (loop
          (sub1 ab-Position)
          (sub1 bis-Position)
          (cdr Individuum-1)
          (cdr Individuum-2)
          (cons (car Individuum-1) akku-1)
          (cons (car Individuum-2) akku-2)))
        ((not (negative? bis-Position))       ;umgekehrt einbauen
         (loop
          (sub1 ab-Position)
          (sub1 bis-Position)
          (cdr Individuum-1)
          (cdr Individuum-2)
          (cons (car Individuum-2) akku-1)
          (cons (car Individuum-1) akku-2)))
        (else                           ;wieder weiter durchgehen
         (loop
          (sub1 ab-Position)
          (sub1 bis-Position)
          (cdr Individuum-1)
          (cdr Individuum-2)
          (cons (car Individuum-1) akku-1)
          (cons (car Individuum-2) akku-2)))))))
