;;;                                                <GA Money.scm>
(require (lib "breakpoint.ss"))
;;; Anwendung des allgemeinen Programmes auf das 
;;; >>>>>     send                              <<<<<
;;; >>>>>  +  more                              <<<<<
;;; >>>>>  = money                    - Problem <<<<<
;;; >>>>>  =======                              <<<<<

;;; Die Erluterungen sind aus diesem Programm weitgehend entnommen.
;;; Siehe dazu den Ordner Genetische Algorithmen.
;;; ===== genetischer-Algorithmus ===================================
(define
  (genetischer-Algorithmus 
   Populations-Groesse
   Anzahl-Schritte
   Mutationswahrscheinlichkeit
   crossing-over-Wahrscheinlichkeit)
  (let
      loop
    ((Population 
      (Bewertung (erzeuge-Population Populations-Groesse)))
     (zaehler Anzahl-Schritte))
    (display zaehler)(display " ")
    (Darstellung (bestes-Individuum Population)) 
    (cond
      ((zero? (cadr (bestes-Individuum Population)))
       (writeln "Lsung:" (bestes-Individuum Population)))
      ((zero? zaehler)
       (Darstellung (bestes-Individuum Population))
        (bestes-Individuum Population))
      (else
       (loop
        (Selektion
         (Kreuzen
          (Mutation Population Mutationswahrscheinlichkeit)
          crossing-over-Wahrscheinlichkeit)
         Populations-Groesse)
        (sub1 zaehler))))))


;;; ===== GA-Aufruf =================================================
; Eine Aufrufhlle fr 'genetischer-Algorithmus'.
(define
  (GA-Aufruf)
  (let
      ((Populations-Groesse
       (begin
         (display "Gib die Zahl der Individuen an [> 0] : ")
         (read)))
       (Anzahl-Schritte
       (begin
         (display "Wieviel Schritte ?             [> 0] : ")
         (read)))
       (Mutationswahrscheinlichkeit
       (begin
         (display "Mutationswahrscheinlichkeit ?  [00 .. 99] : ")
         (read)))
       (crossing-over-Wahrscheinlichkeit
        (begin
          (display 
           "Wahrscheinlichkeit fr crossing-over ? [00 .. 99] : ")
          (read))))
    (genetischer-Algorithmus 
     Populations-Groesse
     Anzahl-Schritte
     Mutationswahrscheinlichkeit
     crossing-over-Wahrscheinlichkeit)))


;;; ===== erzeuge-Population ========================================
; Erzeugt die Grundpopulation fr den genetischen Prozess.
; Parameter sind: Anzahl der Individuen,
; Wert: Eine Liste von Individuen
(define
  (erzeuge-Population Anzahl)
  (let
      loop
    ((Population ())
     (n Anzahl))
    (cond
      ((zero? n)
       Population)
      (else
       (loop
        (cons
         (erzeuge-Individuum)
         Population)
        (sub1 n))))))

;;; ===== aktion? ===================================================
; Hilfsfunktion, die entscheidet ob eine Aktion stattfindet oder
; nicht. Die Entscheidung hngt von der bergebenen ganzzahligen
; Wahrscheinlichkeit p ab.
(define
  (aktion? p)
  (< (random 100) p))

;;; ===== waehle-Positionen =========================================
; Whlt zwei verschiedene Zahlen aus {0 .. n-1} und gib sie als Liste
; zurck. Wird als Hilfsfunktion bentigt.
; Die Zahlen sind stets geordnet: (klein gross)
; Unzulssige Parameter fr n lsen einen Fehler aus!
(define
  (waehle-Positionen n)
  (cond
    ((not (number? n)) (error "keine Zahl !" n))
    ((not (integer? n)) (error "keine ganze Zahl !" n))
    ((< n 2) (error "keine Auswahl mglich !" n))
    (else
     (let
         loop
       ((i (random n))
        (j (random n)))
       (cond
         ((and (= i j) (positive? i)) (list (sub1 i) j))
         ((= i j) (list i (add1 j)))
         ((> i j) (list j i))
         (else
          (list i j)))))))

;;; ===== Mutation ==================================================
; Steuert die Mutation fr die gesamte Population.
; Die bei der Mutation neu erzeugten Individuen werden der Population
; hinzugefgt, dadurch ndert sich die Gesamtzahl. Deren Anpassung
; erfolgt bei der Selektion.
; Parameter sind: Population, Mutationswahrscheinlichkeit
; Wert: mutierte Population.
(define
  (Mutation Population Mutationswahrscheinlichkeit)
  (let
      loop
    ((Population Population)
     (akku ()))
    (cond
      ((null? Population)
       (reverse akku))
      ((aktion? Mutationswahrscheinlichkeit)
       (loop
        (cdr Population)
        (cons
         (bewerte (mutiere (caar Population)))
         (cons
          (car Population)
          akku))))
      (else
       (loop
        (cdr Population)
        (cons
         (car Population)
         akku))))))

;;; ===== Kreuzen ===================================================
; Steuert das Kreuzen fr die gesamte Population.
; Da beim Kreuzen neue Individuen erzeugt und der Population hinzu-
; gefgt werden, ndert sich die Gesamtzahl. Deren Anpassung erfolgt
; bei der Selektion.
; Die zu kreuzenden Individuen werden nach dem Zufallsprinzip ausge-
; whlt.
; Parameter sind: Population, crossing-over-Wahrscheinlichkeit
; Wert: gekreuzte Population.
(define
  (Kreuzen Population crossing-over-Wahrscheinlichkeit)
  (let
      ((n (length Population)))
    (let
        loop
      ((Auswahlen (waehle-Positionen n))
       (i (length Population))                ; Zaehler
       (akku ()))
      (cond
        ((zero? i)
         (append akku Population))
        ((aktion? crossing-over-Wahrscheinlichkeit)
         (let
             ((neue 
               (kreuze
                (car (list-ref Population (car Auswahlen)))
                (car (list-ref Population (cadr Auswahlen))))))
           (loop
            (waehle-Positionen n)
            (sub1 i)
            (cons
             (bewerte (car neue))
             (cons
              (bewerte (cadr neue))
              akku)))))
        (else
         (loop
          (waehle-Positionen n)
          (sub1 i)
          akku))))))

;;; ===== Selektion =================================================
; Es gibt viele verschiedene Selektionsmechanismen. 
; Selektiert wird aus der Menge der Individuen durch direkten
; Vergleich von je zwei Individuen.
(define
  (Selektion Population Gesamtzahl)
  (let
      loop
    ((Auswahlen (waehle-Positionen (length Population)))
     (n (length Population))
     (akku ()))
    (cond
      ((= Gesamtzahl (length akku))
       akku)
      (else
       (loop
        (waehle-Positionen n)
        n
        (cons
         (selektiere
          (list-ref Population (car Auswahlen))
          (list-ref Population (cadr Auswahlen)))
         akku))))))

;;; ===== selektiere ================================================
; Selektiert von zwei Individuen das mit der hheren Bewertungszahl.
(define
  (selektiere Individuum-1 Individuum-2)
  (if
   (besser? Individuum-1 Individuum-2)
   Individuum-1
   Individuum-2))

;;; ===== besser? ===================================================
; Das Prdikat 'besser?' ist das Vergleichsprdikat zur Bewertung und
; ist daher vom speziellen Problem abhngig. Hier wird einfach die
; kleinere (!) Bewertungszahl genommen.
(define
  (besser? Individuum-1 Individuum-2)
   (< (cadr Individuum-1) (cadr Individuum-2)))

;;; ===== Bewertung =================================================
; Fhrt die Bewertung der gesamten Population durch. Dazu wird die 
; Populationsliste in eine Assoziationsliste mit Paaren 
; (Individuum Bewertung) verwandelt.
(define
  (Bewertung Population)
  (let
      loop
    ((Population Population)
     (akku ()))
    (cond
      ((null? Population)
       (reverse akku))
      (else
       (loop
        (cdr Population)
        (cons
         (bewerte (car Population))
         akku))))))

;;; ===== bestes-Individuum =========================================
; Fr die grafische Darstellung, aber auch fr das Ergebnis des GA
; ist es sinnvoll, das 'beste' Individuum bestimmen zu knnen.
(define
  (bestes-Individuum Population)
  (let
      loop
    ((Population (cdr Population))
     (bestes (car Population)))
    (cond
      ((null? Population) bestes)
      ((besser? (car Population) bestes)
       (loop (cdr Population) (car Population)))
      (else
       (loop (cdr Population) bestes)))))

;;; ===== Zufallsgenerator-setzen ===================================
; Setzt den Zufallsgenerator je nach Parameter mit einer "zuflligen"
; Zahl oder der Zahl 123456789.
(define 
  (Zufallsgenerator-setzen Zeit?)
  (random-seed 
   (if
    Zeit?
    (remainder 
     (abs (current-milliseconds))
     1000000000)
    123456789)))


;;; ===== Aufruf ====================================================
(Zufallsgenerator-setzen #f)
(load "GA Money Funktionen.scm")

;;;(genetischer-Algorithmus 
;;;   Populations-Groesse
;;;   Anzahl-Schritte
;;;   Mutationswahrscheinlichkeit
;;;   crossing-over-Wahrscheinlichkeit)

(Zufallsgenerator-setzen #t)
(genetischer-Algorithmus 1000 100 70 50)

;;; ===== Eine Lsung : ======================== 
;;;62 (((y 2) (r 8) (o 0) (m 1) (d 7) (n 6) (e 5) (s 9)) 0) 
;;;9567 
;;;1085 
;;;10652 
;;;
;;;10652 
;;;============================= 
;;;Lsung: (((y 2) (r 8) (o 0) (m 1) (d 7) (n 6) (e 5) (s 9)) 0) 
