;;; <gen Selbst 2.scm>
(require (lib "breakpoint.ss"))
(require (lib "graphics.ss" "graphics"))
; Es werden in einem Quadrat zuf?llig Punkte erzeugt. 
; Dann werden sie einem genetischen Prozess unterworfen,
; bei dem als Bewertungsfunktion die Anzahl der Punkte
; verwendet wird, die in einem gemeinsamen Streifen von Breite eps
; liegen, der indiesem Fall fest mit dem Radius vorgegeben ist.

;;; ===== Zufallsgenerator-setzen ===================================
; Setzt den Zufallsgenerator je nach Parameter mit einer "zuf?lligen"
; Zahl oder der Zahl 123456789.
(define 
  (Zufallsgenerator-setzen Zeit?)
  (random-seed 
   (if
    Zeit?
    (remainder 
     (abs (current-milliseconds))
     1000000000)
    123456789)))

;;; ===== erzeuge-Punkte ============================================
; Erzeugt eine Liste von n zuf?lligen Punkten im Quadrat der Breite
; 'breite'. Alle Koordinaten sind 0 < x,y < breite
(define
  (erzeuge-Punkte n breite Zeit?)
  (Zufallsgenerator-setzen Zeit?)
  (let
      loop
    ((akku ())
     (n n))
    (cond
      ((zero? n) akku)
      (else
       (loop
        (cons (list (add1 (random (sub1 breite)))
                    (add1 (random (sub1 breite)))) akku)
        (sub1 n))))))

;;; ===== Abstand-Ursprung ==========================================
; Berechnet zu jedem Punkt den Abstand vom Ursprung (allerdings das
; Quadrat) und gibt Punkte und Abst?nde als Assoziationsliste
; zur?ck.
(define
  (Abstand Punkte)
  (define
    (d x y)
    (+
     (* x x)
     (* y y)))
  (let
      loop
    ((Punkte Punkte)
     (akku ()))
    (cond
      ((null? Punkte) (reverse akku))
      (else
       (loop 
        (cdr Punkte)
        (cons 
         (list
          (car Punkte)
          (d (caar Punkte) (cadar Punkte)))
         akku))))))
;;; ===== bewerte-Abstaende =========================================
(define
  (bewerte-Abstaende Punkt-Asso eps)
  (let
      loop
    ((Punkt-Asso Punkt-Asso)
     (Radius-Quadrat (* Radius Radius))
     (akku 0))
    (cond
      ((null? (cdr Punkt-Asso))                ; fertig!
       akku)
      ((< (abs (- (cadar Punkt-Asso) Radius-Quadrat)) eps)
       (loop (cdr Punkt-Asso) Radius-Quadrat (add1 akku)))
      (else
       (loop (cdr Punkt-Asso) Radius-Quadrat akku)))))

;;; ===== heize =====================================================
; nimmt eine Punktliste und ver?ndert die Koordinaten mit zuf?lligen,
; aber von dem Parameter Temperatur abh?ngigen Ver?nderungen.
(define
  (heize Punkte Temperatur alle-wieviel?)
  (let
      loop
    ((Punkte Punkte)
     (akku ()))
    (cond
      ((null? Punkte)
       (reverse akku))
      (else
       (loop
        (cdr Punkte)
        (cons
         (if
          (zero? (random alle-wieviel?))          ; wie h?ufig ?
          (list
           (heize-Koordinate  (caar Punkte) Temperatur)
           (heize-Koordinate  (cadar Punkte) Temperatur))
          (car Punkte))
         akku))))))
;;; ===== heize-Koordinate =========================================
(define
  (heize-Koordinate xy Temperatur)
  (abs
   (+ xy (random Temperatur) 
      (quotient (- Temperatur) 2))))
    
;;; ===== ein-Schritt ==============================================
; Nach dem Heizen werden die urspr?ngliche L?sung und die geheizte
; verglichen und liefert die Assoziationsliste zur?ck, welche die
; gr??ere Bewertungszahl hat.
(define
  (ein-Schritt Punkt-Asso Temperatur eps alle-wieviel?)
  (let
      ((alte Punkt-Asso)
       (neue 
        (Abstand 
         (heize
          (Punkte-aus-Asso Punkt-Asso)
          Temperatur
          alle-wieviel?))))
    (if
     (> (bewerte-Abstaende alte eps) (bewerte-Abstaende neue eps))
     alte
     neue)))

;;; ===== Punkte-aus-Asso ===========================================
; Macht aus der Assoziationsliste wieder eine einfache Punktliste. 
(define
  (Punkte-aus-Asso Punkt-Asso)
  (let
      loop
    ((Punkt-Asso Punkt-Asso)
     (akku ()))
    (cond
      ((null? Punkt-Asso) (reverse akku))
      (else
       (loop (cdr Punkt-Asso) (cons (caar Punkt-Asso) akku))))))

;;; ===== viele-Schritte ============================================
; F?hrt das Verfahren f?r n Punkte m Schritte lang durch.
(define
  (viele-Schritte n m Temperatur eps alle-wieviel? Zeit?)
  (let
      loop
    ((Punkt-Asso (zeichne (Abstand (erzeuge-Punkte n Groesse Zeit?))))
     (Temperatur Temperatur) ; kann so ver?ndert werden
     (eps eps)
     (m m))
    (cond
      ((zero? m) Punkt-Asso)
      (else
       (loop 
        (if
         (zero? (modulo m 10)) ; nicht zu oft zeichnen
         (zeichne 
          (ein-Schritt Punkt-Asso Temperatur eps alle-wieviel?))
         (ein-Schritt Punkt-Asso Temperatur eps alle-wieviel?))
        (inexact->exact (round (* Temperatur 1)))
        (inexact->exact (round (* eps 1)))
        (sub1 m))))))
  
;;; ===== Grafik ====================================================
(define Groesse 600)
(define Radius 300)
(open-graphics)
(define w (open-viewport "Gel?nde" Groesse Groesse))
(define (warte) (viewport-flush-input w)(key-value (get-key-press w)))
;;; ===== zeichne ===================================================
; Zeichnet die Punkte aus der Assoziationsliste und gibt sie zur?ck.
(define
  (zeichne Punkt-Asso)
  ((clear-viewport w))
  (let
      loop
    ((P Punkt-Asso))
    (cond
      ((null? P) 
       ;(warte)
       Punkt-Asso)
      (else
       ((draw-pixel w) 
        (make-posn (caaar P) (cadaar P)) (make-rgb 1 0 0))
       (loop (cdr P))))))

;;; ===== Tests =====================================================
;;;(viele-Schritte n m Temperatur eps alle-wieviel? Zeit?)
(viele-Schritte 400 5000 60 1000 10 #f)
;;; ===== Ende der Grafik ===========================================
(warte)
(close-viewport w)
(close-graphics)
 
