;;; <genetische-Algorithmen.scm>
(require (lib "breakpoint.ss"))

; Mutationswahrscheinlichkeit und Selektionswahrscheinlichkeit
; sind positive ganze Zahlen zwischen 0 und 100 ( also % )

;;; ===== Entwurfsentscheidungen ====================================
; Auch wenn der Name 'genetischer-Algorithmus' anderes andeutet, es
; gibt nicht den genetischen Algorithmus, dies ist >>>einer<<< ! 
; Zu den einzelnen Entwurfsentscheidungen siehe auch bei den
; einzelnen Teilprozessen.
; In der Funktion selbst steckt z.B. die Entwurfsentscheidung, den
; Abbruch allein von der Zahl der Schritte abhngig zu machen. Man
; kann auch andere Abbruchkriterien whlen.

; Eine weitere wichtige Entwurfsentscheidung betrifft die Bewertung
; der Individuen in den Populationen. Wenn auch der eigentliche Sinn
; der GA fr die Informatik in der Tatsache steckt, dass die Bewer-
; tungen erheblich 'billiger' sind als das Generieren von Lsungen,
; steckt in der Verwaltung dieser Werte ein abzuwgender Verwaltungs-
; aufwand. Ich habe mich hier daher so entschieden, whrend der
; Bearbeitung die Bewertungen an die Individuen zu koppeln.
; Jedes von ihnen stellt whrend der Bearbeitung daher ein Paar 
; (Individuum Bewertung) dar!
; Das setzt voraus, dass beim jedem Erzeugen eines neuen Individuums
; sofort eine Bewertung durchgefhrt wird.

;;; ===== genetischer-Algorithmus ===================================
(define
  (genetischer-Algorithmus 
   Populations-Groesse
   Anzahl-Schritte
   Mutationswahrscheinlichkeit
   crossing-over-Wahrscheinlichkeit)
  (let
      loop
    ((Population 
      (Bewertung (erzeuge-Population Populations-Groesse)))
     (zaehler Anzahl-Schritte))
    (Darstellung (bestes-Individuum Population)) 
    (writeln "Population : " Population (length Population))
    (cond
      ((zero? zaehler)
        (bestes-Individuum Population))
      (else
       (loop
        (Selektion
         (Kreuzen
          (Mutation Population Mutationswahrscheinlichkeit)
          crossing-over-Wahrscheinlichkeit)
         Populations-Groesse)
        (sub1 zaehler))))))


;;; ===== GA-Aufruf =================================================
; Eine Aufrufhlle fr 'genetischer-Algorithmus', bei der die
; Parameter zur Laufzeit eingegeben werden knnen
(define
  (GA-Aufruf)
  (let
      ((Populations-Groesse
       (begin
         (display "Gib die Zahl der Individuen an [> 0] : ")
         (read)))
       (Anzahl-Schritte
       (begin
         (display "Wieviel Schritte ?             [> 0] : ")
         (read)))
       (Mutationswahrscheinlichkeit
       (begin
         (display "Mutationswahrscheinlichkeit ?  [00 .. 99] : ")
         (read)))
       (crossing-over-Wahrscheinlichkeit
        (begin
          (display 
           "Wahrscheinlichkeit fr crossing-over ? [00 .. 99] : ")
          (read))))
    (genetischer-Algorithmus 
     Populations-Groesse
     Anzahl-Schritte
     Mutationswahrscheinlichkeit
     crossing-over-Wahrscheinlichkeit)))

;;; ===== erzeuge-Population ========================================
; Erzeugt die Grundpopulation fr den genetischen Prozess.
; Parameter sind: Anzahl der Individuen,
; Wert: Eine Liste von Individuen
(define
  (erzeuge-Population Anzahl)
  (let
      loop
    ((Population ())
     (n Anzahl))
    (cond
      ((zero? n)
       Population)
      (else
       (loop
        (cons
         (erzeuge-Individuum)
         Population)
        (sub1 n))))))

;;; ===== erzeuge-Individuum ========================================
; Erzeugt ein Individuum der Population.
; Dieser eigentliche Erzeugungsprozess ist problemabhngig.
(define
  (erzeuge-Individuum)
  (random 100))
;;; ===== aktion? ===================================================
; Hilfsfunktion, die entscheidet ob eine Aktion stattfindet oder
; nicht. Die Entscheidung hngt von der bergebenen ganzzahligen
; Wahrscheinlichkeit p ab.
(define
  (aktion? p)
  (< (random 100) p))

;;; ===== waehle-Positionen =========================================
; Whlt zwei verschiedene Zahlen aus {0 .. n-1} und gib sie als Liste
; zurck. Wird als Hilfsfunktion bentigt.
; Die Zahlen sind stets geordnet: (klein gross)
; Unzulssige Parameter fr n lsen einen Fehler aus!
(define
  (waehle-Positionen n)
  (cond
    ((not (number? n)) (error "keine Zahl !" n))
    ((not (integer? n)) (error "keine ganze Zahl !" n))
    ((< n 2) (error "keine Auswahl mglich !" n))
    (else
     (let
         loop
       ((i (random n))
        (j (random n)))
       (cond
         ((and (= i j) (positive? i)) (list (sub1 i) j))
         ((= i j) (list i (add1 j)))
         ((> i j) (list j i))
         (else
          (list i j)))))))

;;; ===== Mutation ==================================================
; Steuert die Mutation fr die gesamte Population.
; Die bei der Mutation neu erzeugten Individuen werden der Population
; hinzugefgt, dadurch ndert sich die Gesamtzahl. Deren Anpassung
; erfolgt bei der Selektion.
; Parameter sind: Population, Mutationswahrscheinlichkeit
; Wert: mutierte Population.
(define
  (Mutation Population Mutationswahrscheinlichkeit)
  (let
      loop
    ((Population Population)
     (akku ()))
    (cond
      ((null? Population)
       (reverse akku))
      ((aktion? Mutationswahrscheinlichkeit)
       (loop
        (cdr Population)
        (cons
         (bewerte (mutiere (caar Population)))
         (cons
          (car Population)
          akku))))
      (else
       (loop
        (cdr Population)
        (cons
         (car Population)
         akku))))))

;;; ===== mutiere ===================================================
; Mutiert ein einzelnes Individuum.
; Hier muss die Bewertung bei den Individuen fehlen !
; Dieser eigentliche Mutationsprozess ist problemabhngig.
; Individuum -> Individuum
(define
  (mutiere Individuum)
  (quotient (+ (random 100) Individuum) 2))

;;; ===== Kreuzen ===================================================
; Steuert das Kreuzen fr die gesamte Population.
; Da beim Kreuzen neue Individuen erzeugt und der Population hinzu-
; gefgt werden, ndert sich die Gesamtzahl. Deren Anpassung erfolgt
; bei der Selektion.
; Die zu kreuzenden Individuen werden nach dem Zufallsprinzip ausge-
; whlt.
; Parameter sind: Population, crossing-over-Wahrscheinlichkeit
; Wert: gekreuzte Population.
(define
  (Kreuzen Population crossing-over-Wahrscheinlichkeit)
  (let
      ((n (length Population)))
    (let
        loop
      ((Auswahlen (waehle-Positionen n))
       (i (length Population))                ; Zaehler
       (akku ()))
      (cond
        ((zero? i)
         (append akku Population))
        ((aktion? crossing-over-Wahrscheinlichkeit)
         (let
             ((neue 
               (kreuze
                (car (list-ref Population (car Auswahlen)))
                (car (list-ref Population (cadr Auswahlen))))))
           (loop
            (waehle-Positionen n)
            (sub1 i)
            (cons
             (bewerte (car neue))
             (cons
              (bewerte (cadr neue))
              akku)))))
        (else
         (loop
          (waehle-Positionen n)
          (sub1 i)
          akku))))))

;;; ===== kreuze ====================================================
; Kreuzt zwei einzelne Individuen.
; Hier muss die Bewertung bei den Individuen fehlen !
; Dieser eigentliche Kreuzungsprozess ist problemabhngig.
; Individuum x Individuum -> (list Individuum Individuum)
(define
  (kreuze Individuum-1 Individuum-2)
  (list 
    (quotient (+ Individuum-1 Individuum-1 Individuum-1 Individuum-2) 4)
    (quotient (+ Individuum-1 Individuum-2 Individuum-2 Individuum-2) 4)))

;;; ===== Selektion =================================================
; Es gibt viele veschiedene Selektionsmechanismen. Beispielsweise
; kann die Auswahl proportional zur Bewertungsfunktion erfolgen.
; Hier werden die vorhandenen Individuen einfach zufllig ausgewhlt
; und einer paarweisen Auswahl unterzogen: Selektiert wird aus der
; Menge der Individuen durch direkten Vergleich von je zwei
; Individuen. Der Parameter Gesamtzahl gewhrleistet, dass wieder
; die ursprngliche Populationsgrsse hergestellt wird.
; Dazu wird mit der Selektion abgebrochen, wenn eine gengend grosse
; Zahl ausgewhlt wurde.
(define
  (Selektion Population Gesamtzahl)
  (let
      loop
    ((Auswahlen (waehle-Positionen (length Population)))
     (n (length Population))
     (akku ()))
    (cond
      ((= Gesamtzahl (length akku))
       akku)
      (else
       (loop
        (waehle-Positionen n)
        n
        (cons
         (selektiere
          (list-ref Population (car Auswahlen))
          (list-ref Population (cadr Auswahlen)))
         akku))))))

;;; ===== selektiere ================================================
; Selektiert von zwei Individuen das mit der hheren Bewertungszahl.
; Obwohl hier die Bewertungen der Individuen maximal zweimal bestimmt
; werden, kann es sich lohnen, diese Werte zunchst in einer neuen
; Liste oder in der Populationsliste mit abzulegen.
; So kann z.B. auch in der Ausgabe darauf zugegriffen werden.
; siehe-> Bewertung
(define
  (selektiere Individuum-1 Individuum-2)
  (if
   (besser? Individuum-1 Individuum-2)
   Individuum-1
   Individuum-2))

;;; ===== besser? ===================================================
; Das Prdikat 'besser?' ist das Vergleichsprdikat zur Bewertung und
; ist daher vom speziellen Problem abhngig. Hier wird einfach die
; hhere Bewertungszahl genommen.
(define
  (besser? Individuum-1 Individuum-2)
   (> (cadr Individuum-1) (cadr Individuum-2)))

;;; ===== Bewertung =================================================
; Fhrt die Bewertung der gesamten Population durch. Dazu wird die 
; Populationsliste in eine Assoziationsliste mit Paaren 
; (Individuum Bewertung) verwandelt.
(define
  (Bewertung Population)
  (let
      loop
    ((Population Population)
     (akku ()))
    (cond
      ((null? Population)
       (reverse akku))
      (else
       (loop
        (cdr Population)
        (cons
         (bewerte (car Population))
         akku))))))

;;; ===== bewerte ===================================================
; Bewertet ein Individuum mit einer Bewertungsziffer.
; Liefert eine Liste aus Individuum und Berteungszahl zurck
; Die Bewertungsfunktion hngt vom speziellen Problem ab.
(define
  (bewerte Individuum)
  (list Individuum (* 10 Individuum)))

;;; ===== bestes-Individuum =========================================
; Fr die grafische Darstellung, aber auch fr das Ergebnis des GA
; ist es sinnvoll, das 'beste' Individuum bestimmen zu knnen.
(define
  (bestes-Individuum Population)
  (let
      loop
    ((Population Population)
     (bestes (car Population)))
    (cond
      ((null? Population) bestes)
      ((besser? (car Population) bestes)
       (loop (cdr Population) (car Population)))
      (else
       (loop (cdr Population) bestes)))))

;;; ===== Darstellung ===============================================
; Die Darstellung der 'besten' Lsung ist problemabhngig.
(define
  (Darstellung Individuum)
  (display Individuum)
  (newline))

;;; ===== Zufallsgenerator-setzen ===================================
; Setzt den Zufallsgenerator je nach Parameter mit einer "zuflligen"
; Zahl oder der Zahl 123456789.
(define 
  (Zufallsgenerator-setzen Zeit?)
  (random-seed 
   (if
    Zeit?
    (remainder 
     (abs (current-milliseconds))
     1000000000)
    123456789)))


;;; ===== Aufruf ====================================================
(Zufallsgenerator-setzen #t)
(genetischer-Algorithmus 5 10 20 30)
