;;; <Missio Kani.scm>
(require (lib "breakpoint.ss"))
;;; >>>>> Vorbemerkungen <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
;;; Bei der vorgegebenen Problemstellung muss vor und nach einem
;;; Transport die Summe der Missionare und der Kannibalen jeweils
;;; konstant 3 sein, so dass nur eine Seite verwaltet werden muss.

;;; Es werden beim Generieren der Nachfolger zunchst alle Zustnde
;;; generiert und dann die unzulssigen aussortiert.

(define Start-Zustand (list 3 3 'links))
(define Ziel-Zustand (list 0 0 'rechts))
;;; ===== Zugriffsfunktionen ========================================
; Zugriffsfunktionen fr die einzelnen Komponenten.
(define (Liegeplatz Zustand) (caddr Zustand))
(define (M-links Zustand) (car Zustand))
(define (K-links Zustand) (cadr Zustand))
(define (M-rechts Zustand) (- 3 (car Zustand)))
(define (K-rechts Zustand) (- 3 (cadr Zustand)))

;;; ===== Bedingungen ===============================================
; Es treten Bedingungen fr links und fr rechts auf. Ist das Boot
; links, mssen rechts die Bedingungen erfllt sein und umgekehrt.
; Die Prffunktionen liefern jeweils eine Liste mit dem Zustand.
;;; ----- links-OK --------------------------------------------------
(define
  (links-OK Zustand)
  (cond
    ((negative? (M-links Zustand)) ())
    ((negative? (K-links Zustand)) ())
    ((and
      (positive? (M-links Zustand))
      (> (K-links Zustand) (M-links Zustand))) ())
    (else (list Zustand))))

;;; ----- rechts-OK -------------------------------------------------
(define
  (rechts-OK Zustand)
  (cond
    ((negative? (M-rechts Zustand)) ())
    ((negative? (K-rechts Zustand)) ())
    ((and
      (positive? (M-rechts Zustand))
      (> (K-rechts Zustand) (M-rechts Zustand))) ())
    (else (list Zustand))))

;;; ===== Boote =====================================================
; Es werden die mglichen Boote definiert. Sie bekommen den aktuellen
; Zustand bergeben und geben eine Liste mit dem Zielzustand zurck.
;;; ----- Boot-1 ----------------------------------------------------
; Zwei Missionare nach rechts.
(define
  (Boot-1 Zustand)
  (list 
   (- (M-links Zustand) 2)
   (K-links Zustand)
   'rechts))

;;; ----- Boot-2 ----------------------------------------------------
; Zwei Kannibalen nach rechts.
(define
  (Boot-2 Zustand)
  (list 
   (M-links Zustand)
   (- (K-links Zustand) 2)
   'rechts))

;;; ----- Boot-3 ----------------------------------------------------
; Je einen nach rechts.
(define
  (Boot-3 Zustand)
  (list 
   (sub1 (M-links Zustand))
   (sub1 (K-links Zustand))
  'rechts))

;;; ----- Boot-4 ----------------------------------------------------
; Einen Missionar nach links.
(define
  (Boot-4 Zustand)
  (list 
   (add1 (M-links Zustand))
   (K-links Zustand)
   'links))

;;; ----- Boot-5 ----------------------------------------------------
; Einen Kannibalen nach links.
(define
  (Boot-5 Zustand)
  (list 
   (M-links Zustand)
   (add1 (K-links Zustand))
   'links))

;;; ===== Nachfolger ================================================
; Generiert alle zulssigen Nachfolger eines Zustandes.
(define
  (Nachfolger Zustand)
;  (bkpt (links-OK (Boot-1 Zustand)) Zustand)
  (if
   (equal? (Liegeplatz Zustand) 'links)
   (append
    (links-OK (Boot-1 Zustand))
    (links-OK (Boot-2 Zustand))
    (links-OK (Boot-3 Zustand)))
   (append
    (rechts-OK (Boot-4 Zustand))
    (rechts-OK (Boot-5 Zustand)))))

;;; ===== Tiefensuche ===============================================
; Der Suchraum entfllt.
(define
  (Tiefensuche Start Ziel)
  
  (let
      t-s
    ((Alternativen (Nachfolger Start))
     (besucht (list Start)))
    (cond
      ((null? Alternativen) #f)
      
      ((equal? (car Alternativen) Ziel)
       (writeln (reverse (cons (car Alternativen) besucht)))
       (newline)
       #t)
      
      ((member (car Alternativen) besucht)
       (t-s
        (cdr Alternativen)
        besucht))
      
      (else
       (t-s
        (Nachfolger (car Alternativen) )
        (cons (car Alternativen) besucht))
      
       (t-s
        (cdr Alternativen)
        besucht)))))

    
;;; ----- Tests -----------------------------------------------------
;(Nachfolger (list 3 3 'links))
;(Nachfolger (list 3 2 'links))
;(Nachfolger (list 3 1 'links))
;(Nachfolger (list 3 0 'links))
;(Nachfolger (list 2 3 'links))
;(Nachfolger (list 2 2 'links))
;(Nachfolger (list 2 1 'links))
;(Nachfolger (list 2 0 'links))
;(Nachfolger (list 1 3 'links))
;(Nachfolger (list 1 2 'links))
;(Nachfolger (list 1 1 'links))
;(Nachfolger (list 1 0 'links))
;
;(Nachfolger (list 3 3 'rechts))
;(Nachfolger (list 3 2 'rechts))
;(Nachfolger (list 3 1 'rechts))
;(Nachfolger (list 3 0 'rechts))
;(Nachfolger (list 2 3 'rechts))
;(Nachfolger (list 2 2 'rechts))
;(Nachfolger (list 2 1 'rechts))
;(Nachfolger (list 2 0 'rechts))
;(Nachfolger (list 1 3 'rechts))
;(Nachfolger (list 1 2 'rechts))
;(Nachfolger (list 1 1 'rechts))
;(Nachfolger (list 1 0 'rechts))
(Tiefensuche Start-Zustand Ziel-Zustand)
