;;; <Breitensuche.scm>

;;; ===== Nachfolger =========================================
;;; Bestimmt zu einem Knoten die Nachfolger aus dem Graphen.
(define
   (Nachfolger Element graph)
   (cadr (assoc Element graph)))

;;;===========================================================
(define
   graph
   '( (1 (2 4))      (2 (1 3))      (3 (2 6))
      (4 (1 5 7))    (5 (4 6 8))    (6 (3 5 9))
      (7 (4))        (8 (5 9))      (9 (8 6))))
; Der Graph enth?lt also einen Zyklus !!!

;;;===========================================================
;;; Labyrinth dazu :
;;;                ???????????
;;;                ?1 - 2 - 3?
;;;                ?I ????? I?
;;;                ?4 - 5 - 6?
;;;                ?I ? I ? I?
;;;                ?7 ? 8 - 9?
;;;                ???????????

;;; ===== expandiere-Knoten ==================================
; Der erste Weg wird aus der Wegeliste entnommen, aus dem
; Graphen die Liste der Wege geholt, zu denen eine Verbindung
; f?hrt und es werden die gefundenen neuen Wege an die
; wege-liste angeh?ngt.
; wege-liste --> '((3 2 1) (5 4 1) ...)
; = Prinzip der Warteschlange.    
(define
  (expandiere-Knoten wege-liste graph)
  (let
      ((knoten (caar wege-liste))                        ;{1} 
       (der-weg (car wege-liste)))
    (let
        loop
      ((die-Nachfolger (Nachfolger knoten graph))
       (akku ()))
      (cond
        ((null? die-Nachfolger)                          ;{2}
         (append (cdr wege-liste) akku))
        ((member (car die-Nachfolger) der-weg)           ;{3}
         (loop (cdr die-Nachfolger) akku))
        (else                                            ;{4}
         (loop 
          (cdr die-Nachfolger)
          (cons
           (cons
            (car die-Nachfolger)
            der-weg)
           akku)))))))
;{1} knoten ist der erste Koten aus dem ersten Weg, der in der
;    wege-liste abgelegt ist.
;{2} Wenn es ?berhaupt noch Nachfolger des aktuellen Knoten
;    gab, sind sie jetzt dem Weg jeweils hinzugef?gt worden,
;    in akku abgelegt und m?ssen nun >>>hinten<<< an die wege-
;    liste angeh?ngt werden.
;{3} Ist der aktuelle Nachfolgeknoten schon in der wege-liste
;    enthalten, ist ein Zyklus gefunden und der neue Weg wird
;    nicht in akku ?bernommen.
;{4} Anderenfalls wird der um den Knoten verl?ngerte Weg dem
;    akku hinzugef?gt.
;;; ===== Breitensuche =======================================
; Die Funktion Breitensuche selbst muss nun nur noch die 
; Steuerung der o.a. Funktion erledigen.
(define
  (Breitensuche Start Ziel graph)
  (let
      loop
    ((wege-liste (list (list Start))))
    (cond
      ((null? wege-liste) #f)
      ((member Ziel (Nachfolger (caar wege-liste) graph))
       (reverse (cons Ziel (car wege-liste))))
      (else
       (loop
        (expandiere-Knoten wege-liste graph))))))
;;; ===== Testaufrufe ========================================
;(expandiere-Knoten  '((3 2 1) (5 4 1) (7 4 1)) graph) 
;(expandiere-Knoten  '((5 4 1) (7 4 1) (6 3 2 1)) graph)
;(load "graph.scm")
(Breitensuche 1 9 graph)
