;;;                                   "einfache-tiefensuche-entwicklung.scm"
;;; Demonstration der Entwicklung des Programmes zur Tiefensuche
;;; ----- die Kantenliste -----
; Wir merken sie uns einmal, damit wir sie nicht jedesmal komplett hinschreiben müssen.
; Das Hochkomma verhindert die Auswertung. 
(define
  Kantenliste
  '((0 (3))
    (1 (2 4))
    (2 (1))
    (3 (0 4))
    (4 (1 3 5 7))
    (5 (4 8))
    (6 (9))
    (7 (4 8 10))
    (8 (5 7))
    (9 (6 10 12))
    (10 (7 9 11 13))
    (11 (10))
    (12 (9))
    (13 (10 14))
    (14 (13)))
  )
;;; ----- Funktionskopf -----
; 1. Schritt: Bestimmung des Funktionskopfes
(define
  (Tiefensuche Kantenliste Zielknoten Alternativen)
  #f)
; Testaufruf:
(Tiefensuche '() 10 '())

;;; ----- am Ziel? -----
; 2. Schritt: ich bin am Ziel!
; Das tritt nur dann ein, wenn die erste aktuelle Alternative gleich dem Zielknoten ist.
; Hier ist also eine Bedingung zu überprüfen. --> if
(define
  (Tiefensuche Kantenliste Zielknoten Alternativen)
  (if
   (equal? Zielknoten (car Alternativen))         ; Bedingung
   Zielknoten                                     ; ja-Fall
   #f)                                            ; nein-Fall
  )
; Testaufrufe:
(Tiefensuche '() 10 '(10 14))
(Tiefensuche '() 10 '(1 3 5 7))

;;; ----- am "Ende", aber nicht am Ziel -----
; 3. Schritt: es geht nicht weiter!
; Auch mit dieser Abbruchbedingung muss die Funktion umgehen können.
; Das tritt nur dann ein, wenn es keine Alternative mehr gibt.
; Hier ist also wieder eine Bedingung zu überprüfen. Wir könnten ein weiters if
; verwenden, da Scheme aber eine sehr komfortable Mehrfachverzweigung bereitstellt,
; verwenden wir diese: --> cond
(define
  (Tiefensuche Kantenliste Zielknoten Alternativen)  
  (cond
    ; die neue Bedingung muss zuerst geprüft werden! Warum?
    [(null? Alternativen)                           ; 1.Bedingung
     'Wert-noch-unklar]                             ; ja-Fall
    [(equal? Zielknoten (car Alternativen))         ; 2.Bedingung
     Zielknoten]                                    ; ja-Fall
    [else                                           ; und sonst?
     #f]                                            ; der "sonst"-Fall
    )
  )
; Testaufrufe:
(Tiefensuche '() 10 '(10 14))
(Tiefensuche '() 10 '(1 3 5 7))
(Tiefensuche '() 10 '())
;;; ----- Rekusionssteuerung -----
; 4. Schritt: Rekusionssteuerung
; Es gibt keine weiteren besonderen Fälle.
; Die Funktion muss nun um die Steuerung der Rekursion erweitert werden.
; Sie muss ...
;     ... versuchen vom aktullen Knoten aus in der Tiefe eine Lösung zu finden und(!)
;     ... wenn das nicht erfolgreich ist, den aktuellen Knoten verwerfen und es mit
;         Alternativen versuchen.
(define
  (Tiefensuche Kantenliste Zielknoten Alternativen)  
  (cond
    ; die neue Bedingung muss zuerst geprüft werden! Warum?
    [(null? Alternativen)                             ; 1.Bedingung
     #f]                                              ; (s.u.)
    [(equal? Zielknoten (car Alternativen))           ; 2.Bedingung
     Zielknoten]                                      ; ja-Fall
    [(Tiefensuche                                     ; Schritt in die Tiefe ...
      Kantenliste
      Zielknoten
      (Nachfolger (car Alternativen) Kantenliste))    ; ... mit den Nachfolgern
     ;;; Deren Bestimmung überlassen wir einer gesonderten Funktion (s.u.)
     ]
    ;;; Eine der besonderen Möglichkeiten von cond: Bedingung=Wert
    ;;; Die Bedingung hat einen Wert, der zurückgegeben werden kann.
    ;;; Bei erfolglosem Versuch brauchen wir dann hier den Wert #f,
    ;;; damit zum else gegangen wird.
    [else                                            ; und sonst?
     (Tiefensuche
      Kantenliste
      Zielknoten
      (cdr Alternativen))                 ; bei der naechsten Alternative weiter!
     ]
    )
  )
; Bevor wir zu den Testaufrufen kommen, muessen wir daher erst die Nachfolger-Funktion schreiben.
;;; ----- Nachfolger -----
; Sie greift auf die Kantenliste zu, welche die Form einer Assoziationsliste hat.
; Dafür gibt es eine Standardfunktion von Scheme: assoc.
(define
  (Nachfolger Element Liste)
  (car (cdr (assoc Element Liste))))
; Nun können wir die Testaufrufe starten:
(Tiefensuche Kantenliste 10 '(10 14))
;(Tiefensuche Kantenliste 10 '(1 3 5 7))       ;vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv;
;(Tiefensuche Kantenliste 10 '())              ;vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv;
; (Tiefensuche Kantenliste 10 '(Nachfolger 0)) ;vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv;
; Allerdings führen sie zu einem frustrierenden Ergebnis, weil die Aufrufe schnell in einer 
; endlosen Schachtelung versacken: Beispielesweise führt der letzte zu einem Hin- und Herspringen 
; zwischen den Knoten 0 und 3.
; Unsere Funktion erkennt das nicht!
