;;;                              "neue-schematische-Tiefensuche.scm"
(require (lib "breakpoint.ss"))
;;; ===== schematische-Tiefensuche ==================================
;;; Die Tiefensuche ist zunaechst abstrakt definiert. Erst die 
;;; jeweilige Implementation der Prozedur Nachfolger und die
;;; Behandlung der Alternativen als Liste haengt von der Art
;;; der Repraesentation des Suchraumes ab.
;;; Es wird hier davon ausgegangen, dass dieser bezogen auf einen
;;; Graphen gesucht wird, dessen Kanten in einer Assoziationsliste
;;; vorliegen.
;;; Muster :
;;; ( (1.Knoten (1.Nachfolger 2.Nachfolger ...))
;;;   (2.Knoten (1.Nachfolger 2.Nachfolger ...))
;;;    ... )
;;;
;;; Die folgende Variante ermoeglicht nur eine Suche nach der
;;; ersten Loesung.
;;; Zyklen werden abgefangen:
;;; Damit keine Zyklen entstehen, muessen die schon besuchten
;;; Ecken in einer Liste "besucht" gefuehrt werden.
;;; ===== Tiefensuche ===============================================
; fuehrt eine Tiefensuche mit backtracking unter Verwendung 
; des Systemstacks aus. 
; Knoten x Knoten x Assoziationsliste --> boolean
; Die Ausgabe des Weges erfolgt aus der Liste besucht, wenn
; eine Suche erfolgreich war.
; Durch die lokale let-Schleife wird das bergeben von Start, Ziel
; und Suchraum vermieden.

(define
  (Tiefensuche Start Ziel Graph)
  
  (let
      t-s
    ((Alternativen (Nachfolger (list Start) (cadr (assoc Start Graph)))))
    ;(bkpt Alternativen)
    (cond
      ((null? Alternativen) #f)                      ; [Anm.1]
      
      ((eq? (caar Alternativen) Ziel)                ; [Anm.3]
       ;(writeln (reverse (car Alternativen)))
       (reverse (car Alternativen)))
      
      ((t-s                                          ; [Anm.4]:
        (Nachfolger
         (car Alternativen)
         (cadr (assoc (caar Alternativen) Graph)))))
      
      (else
       (t-s                                           ; [Anm.5]
        (cdr Alternativen))))))

; [Anm.1]: Wenn auf einer Rekursionstiefe alle Alternativen 
;   untersucht wurden, wird mit dem Wert "#f" zur aufrufenden
;   Ebene zurueckgesprungen.
; [Anm.2]: Wenn der aktuelle Knoten schon besucht wurde, wird
;   versucht, mit dem naechsten Knoten der selben Ebene weiter
;   zu arbeiten.
;   Dies ist also der erste Fall einer Rekursion in der Breite.
; [Anm.3]: Wenn das Ziel erreicht wurde, wird die Liste der
;   besuchten Knoten in der richtigen Reihenfolge ausgegeben und
;   und auch zur aufrufenden Ebene zurueckgegeben
; [Anm.4]: Hier erfolgt der eigentliche Abstieg in der Rekursion.
; [Anm.5]: Wenn die Rekursion in der Tiefe erfolglos war, muss
;   noch die Rekursion in der Breite versucht werden.
;   ( siehe auch [Anm.2] )

;;; ===== Nachfolger =========================================
; Bestimmt hier die Nachfolger im Suchraum.
(define
  (Nachfolger weg nachbarn)
  (cond
    ((null? nachbarn) '())
    ((member (car nachbarn) weg)           ; [Anm.2]
     (Nachfolger weg (cdr nachbarn)))
    (else
     (cons
      (cons (car nachbarn) weg)
      (Nachfolger weg (cdr nachbarn))))))

;;;===========================================================
(define
   Graph
   '( (1 (2 4))
      (2 (1 3))
      (3 (2 6))
      (4 (1 5 7))
      (5 (4 6 8))
      (6 (3 5 9))
      (7 (4))
      (8 (5 9))
      (9 (8 6)))) ; enthlt also einen Zyklus !!!

;;;==============================================================
;;; Labyrinth dazu :
;;;                
;;;                1 - 2 - 3
;;;                I  I
;;;                4 - 5 - 6
;;;                I  I  I
;;;                7  8 - 9
;;;                
;;;==============================================================

;(load "grosser-graph.scm")
(Tiefensuche 1 2 Graph)
(Tiefensuche 1 3 Graph)
(Tiefensuche 1 4 Graph)
(Tiefensuche 1 5 Graph)
(Tiefensuche 1 6 Graph)
(Tiefensuche 1 7 Graph)
(Tiefensuche 1 8 Graph)
(Tiefensuche 1 9 Graph)

