;;;                            <greedy bewerteter Graph.scm>
; Ben?tigt "minimal mit greedy Grafik.scm"
; Der Graph besteht in diesem Beispiel aus folgender Figur:
; 1           2
;       5
;   8       6
;       7
; 4           3
; Dabei haben die Punkte die Koordinaten
(define
  Koordinaten
  '(
    (1 (10 10))
    (2 (390 11))
    (3 (390 390))
    (4 (12 390))
    (5 (201 99))
    (6 (303 199))
    (7 (202 298))
    (8 (104 201))))
; Es sollen alle Verbindungen m?glich sein.
;;; ===== Abstand ===================================================
; Bewertung durch den geometrischen Abstand.
(define
  (Abstand P-1 P-2)
  (sqrt (+ (sqr (- (car P-1) (car P-2)))
           (sqr (- (cadr P-1) (cadr P-2))))))
(define (sqr x) (* x x))
;;; ===== Graph =====================================================
; Bewerteter Graph.  
(define
  Graph
  (let loop
    ((Punkt 8)
     (alle-Punkte Koordinaten)
     (akku ())
     (temp ()))
    (cond
      ((zero? Punkt)
       akku)
      ((null? alle-Punkte)
       (loop 
        (sub1 Punkt)
        Koordinaten
        (cons (list Punkt (reverse temp)) akku)
        ()))
      ((equal? Punkt (caar alle-Punkte))
       (loop Punkt (cdr alle-Punkte) akku temp))
      (else
       (loop
        Punkt
        (cdr alle-Punkte)
        akku
        (cons (list (caar alle-Punkte) 
                    (Abstand
                     (cadr (assoc Punkt Koordinaten))
                     (cadar alle-Punkte)))
              temp))))))

;;; ===== Nachfolger =========================================
; Anders als bei der einfachen Tiefensuche muss die Nachfolger-
; funktion die Kanten komplett ?bergeben, also auch mit dem 
; Ausgangsknoten.
(define
  (Nachfolger Knoten Graph)
  (let
      loop
    ((liste (cadr (assoc Knoten Graph)))
     (akku ()))
    (cond
      ((null? liste) akku)
      (else
       (loop
        (cdr liste)
        (cons
         (cons
          Knoten
          (car liste))
         akku))))))

;;; ===== laenger? ==========================================
; Pr?dikat zum Sortieren
(define
  (laenger? erstes zweites)
  (> (caddr erstes) (caddr zweites)))

;;; ===== ordne ==============================================
; ordnet ?bergebene Kanten der Laenge nach durch lineares
; Einf?gen.
(define
  (ordne Kanten)
  (cond
    ((null? Kanten) ())
    (else
     (fuege-ein (car Kanten) (ordne (cdr Kanten))))))

;;; ----- fuege-ein ------------------------------------------
; Hilfprozedur zum linearen Einf?gen in eine sortierte Liste.
(define
  (fuege-ein Element aufwaerts-sortierte-Liste)
  (cond
    ((null? aufwaerts-sortierte-Liste) (list Element))
    ((laenger? Element (car aufwaerts-sortierte-Liste))
     (cons 
      (car aufwaerts-sortierte-Liste)
      (fuege-ein Element (cdr aufwaerts-sortierte-Liste))))
    (else
     (cons Element aufwaerts-sortierte-Liste))))

;;; ===== greedy =============================================
; Die benachbarten Kanten werden in einer Priorit?ts - Warte-
; schlange gef?hrt: In sie werden die Elemente nach ihren 
; Bewertungen sortiert eingef?gt.
(define
  (greedy Startknoten Graph)
  (let
     ((laenge (length Graph)))
      (let
          loop
        ((Prio-Schlange 
          (ordne (Nachfolger Startknoten Graph)))
         (besucht (list Startknoten))
         (beteiligte-Kanten ()))
;        (display Prio-Schlange)(newline)
;        (display besucht)(newline)
;        (display beteiligte-Kanten)(newline)(newline)
        (cond
          ((= laenge (length besucht))                    ;{1}
           beteiligte-Kanten)
          ((member (Zielknoten (car Prio-Schlange)) besucht)
           (loop                                          ;{2}
            (cdr Prio-Schlange)
            besucht
            beteiligte-Kanten))
          (else
           (loop                                          ;{3}
            (ordne 
             (append
              (Nachfolger (Zielknoten (car Prio-Schlange)) Graph) 
              (cdr Prio-Schlange)))
            (cons (Zielknoten (car Prio-Schlange))
                  besucht)
            (cons (car Prio-Schlange) beteiligte-Kanten)))))))

;{1} Wenn alle Knoten in der besucht - Liste enthalten sind, ist
;    ihre L?nge gleich der L?nge des Graphen und das Problem ist
;    gel?st.
;{2} F?hrt der erste Weg zu einem Knoten, der schon in der
;    besucht - Liste enthalten ist, muss er aus der Schlange
;    entfernt werden.
;{3} Anderenfalls wird ein Schritt ausgef?hrt. Die neuen Randwege
;    werden in die Priorit?ts - Warteschlange eingef?gt, aus der
;    der erste weg entfernt wurde, der Zielknoten wird der
;    besucht - Liste hinzugef?gt und die Kante der Kantenliste.

;;; ===== Zielknoten =========================================
; Hilfsprozedur zur Bestimmung des Zielknotens.
(define
  (Zielknoten Kante)
  (cadr Kante))
    
;;; ===== Test: ==============================================
(display "Grafik <#t / #f> ?  --> ")
(cond
 ((equal? (read) #t)
  (load "minimal mit greedy Grafik.scm")
  (zeichne (greedy 1 Graph) Koordinaten))
 (else
  (greedy 1 Graph)))
